import sys

from inflection import pluralize

from .builder import tidy_sql
from .enums import GrantType, ResourceType
from .identifiers import FQN, URN
from .props import Props
from .resource_name import ResourceName

__this__ = sys.modules[__name__]


def fqn_to_sql(fqn: FQN):
    database = f"{ResourceName(fqn.database)}." if fqn.database else ""
    schema = f"{ResourceName(fqn.schema)}." if fqn.schema else ""
    name = ResourceName(fqn.name)
    return f"{database}{schema}{name}"


################ Create functions


def create_resource(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    return getattr(__this__, f"create_{urn.resource_label}", create__default)(urn, data, props, if_not_exists)


def create__default(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    return tidy_sql(
        "CREATE",
        urn.resource_type,
        "IF NOT EXISTS" if if_not_exists else "",
        fqn_to_sql(urn.fqn),
        props.render(data),
    )


def create_account_parameter(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    value = data["value"]
    if isinstance(value, str):
        value = f"'{value}'"
    return tidy_sql(
        "ALTER",
        "ACCOUNT",
        "SET",
        urn.fqn.name,
        "=",
        value,
    )


def create_aggregation_policy(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    return tidy_sql(
        "CREATE",
        "AGGREGATION POLICY",
        "IF NOT EXISTS" if if_not_exists else "",
        fqn_to_sql(urn.fqn),
        "AS () RETURNS AGGREGATION_CONSTRAINT",
        props.render(data),
    )


def create_database(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    data = data.copy()
    transient = data.pop("transient", None)
    return tidy_sql(
        "CREATE",
        "TRANSIENT" if transient else "",
        urn.resource_type,
        "IF NOT EXISTS" if if_not_exists else "",
        urn.fqn,
        props.render(data),
    )


def create_database_role_grant(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    if data["to_role"] is not None:
        to = data["to_role"]
        to_type = "ROLE"
    else:
        to = data["to_database_role"]
        to_type = "DATABASE ROLE"

    return tidy_sql(
        "GRANT",
        "DATABASE ROLE",
        data["database_role"],
        "TO",
        to_type,
        to,
    )


def create_function(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    db = f"{urn.fqn.database}." if urn.fqn.database else ""
    schema = f"{urn.fqn.schema}." if urn.fqn.schema else ""
    name = f"{db}{schema}{data['name']}"
    return tidy_sql(
        "CREATE",
        urn.resource_type,
        "IF NOT EXISTS" if if_not_exists else "",
        name,
        props.render(data),
    )


def create_hybrid_table(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    data = data.copy()
    columns = data.pop("columns", [])
    indexes = data.pop("indexes", None)

    # Render columns
    col_strs = []
    for column in columns:
        name = column["name"]
        data_type = str(column["data_type"])
        not_null = " NOT NULL" if column.get("not_null") else ""

        if isinstance(column.get("default"), str):
            default = f" DEFAULT '{column['default']}'"
        elif column.get("default") is not None:
            default = f" DEFAULT {column['default']}"
        else:
            default = ""

        comment = f" COMMENT '{column['comment']}'" if column.get("comment") else ""
        constraint = f" {column['constraint']}" if column.get("constraint") else ""
        col_strs.append(f"{name} {data_type}{not_null}{default}{comment}{constraint}")

    # Render indexes
    if indexes:
        for idx in indexes:
            idx_name = idx["name"]
            idx_columns = ", ".join(idx["columns"])
            idx_str = f"INDEX {idx_name} ({idx_columns})"
            if idx.get("include"):
                idx_include = ", ".join(idx["include"])
                idx_str += f" INCLUDE ({idx_include})"
            col_strs.append(idx_str)

    schema_sql = f"({', '.join(col_strs)})"

    return tidy_sql(
        "CREATE",
        urn.resource_type,
        "IF NOT EXISTS" if if_not_exists else "",
        fqn_to_sql(urn.fqn),
        schema_sql,
        props.render(data),
    )


def create_grant(urn: URN, data: dict, props: Props, if_not_exists: bool):
    on_type = data["on_type"]
    if "INTEGRATION" in str(on_type):
        on_type = "INTEGRATION"
    elif on_type == "ACCOUNT":
        on_type = ""
    if data["grant_type"] == GrantType.FUTURE:
        items_type = data["items_type"]
        if "INTEGRATION" in items_type:
            items_type = "INTEGRATION"
        return tidy_sql(
            "GRANT",
            data["priv"],
            "ON FUTURE",
            pluralize(items_type).upper(),
            "IN",
            data["on_type"],
            data["on"],
            "TO",
            data["to_type"],
            data["to"],
            "WITH GRANT OPTION" if data["grant_option"] else "",
        )
    elif data["grant_type"] == GrantType.ALL:
        items_type = data["items_type"]
        if "INTEGRATION" in items_type:
            items_type = "INTEGRATION"
        return tidy_sql(
            "GRANT",
            data["priv"],
            "ON ALL",
            pluralize(items_type),
            "IN",
            data["on_type"],
            data["on"],
            "TO",
            data["to_type"],
            data["to"],
            "WITH GRANT OPTION" if data["grant_option"] else "",
        )
    return tidy_sql(
        "GRANT",
        data["priv"],
        "ON",
        on_type,
        data["on"],
        "TO",
        data["to_type"],
        data["to"],
        "WITH GRANT OPTION" if data["grant_option"] else "",
    )


def create_masking_policy(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    return tidy_sql(
        "CREATE",
        urn.resource_type,
        "IF NOT EXISTS" if if_not_exists else "",
        urn.fqn,
        "AS",
        props.render(data),
    )


def create_procedure(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    if if_not_exists:
        raise Exception("IF NOT EXISTS not supported for CREATE PROCEDURE")
    db = f"{urn.fqn.database}." if urn.fqn.database else ""
    schema = f"{urn.fqn.schema}." if urn.fqn.schema else ""
    name = f"{db}{schema}{urn.fqn.name}"
    return tidy_sql(
        "CREATE",
        urn.resource_type,
        name,
        props.render(data),
    )


def create_row_access_policy(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    return tidy_sql(
        "CREATE",
        urn.resource_type,
        "IF NOT EXISTS" if if_not_exists else "",
        urn.fqn,
        "AS",
        props.render(data),
    )


def create_role_grant(urn: URN, data: dict, props: Props, if_not_exists: bool = False):
    if data["to_role"] is not None:
        to = data["to_role"]
        to_type = "ROLE"
    else:
        to = data["to_user"]
        to_type = "USER"
    return tidy_sql(
        "GRANT",
        "ROLE",
        data["role"],
        "TO",
        to_type,
        to,
    )


def create_scanner_package(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    package_name = f"'{urn.fqn.name}'"
    return tidy_sql(
        "CALL SNOWFLAKE.TRUST_CENTER.SET_CONFIGURATION(",
        "'ENABLED',",
        "'TRUE',",
        package_name,
        ")",
    )


def create_schema(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    data = data.copy()
    transient = data.pop("transient", None)
    return tidy_sql(
        "CREATE",
        "TRANSIENT" if transient else "",
        urn.resource_type,
        "IF NOT EXISTS" if if_not_exists else "",
        urn.fqn,
        props.render(data),
    )


def create_table(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    data = data.copy()
    transient = data.pop("transient", None)
    return tidy_sql(
        "CREATE",
        "TRANSIENT" if transient else "",
        urn.resource_type,
        "IF NOT EXISTS" if if_not_exists else "",
        urn.fqn,
        props.render(data),
    )


def create_tag_reference(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    tags_sql = ", ".join([f"{k}='{v}'" for k, v in data["tags"].items()])
    return tidy_sql(
        "ALTER",
        data["object_domain"],
        data["object_name"],
        "SET TAG",
        tags_sql,
    )


def create_tag_masking_policy_reference(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    return tidy_sql(
        "ALTER TAG",
        data["tag_name"],
        "SET MASKING POLICY",
        data["masking_policy_name"],
    )


def create_view(urn: URN, data: dict, props: Props, if_not_exists: bool = False) -> str:
    data = data.copy()
    secure = data.pop("secure", None)
    volatile = data.pop("volatile", None)
    recursive = data.pop("recursive", None)
    return tidy_sql(
        "CREATE",
        "SECURE" if secure else "",
        "VOLATILE" if volatile else "",
        "RECURSIVE" if recursive else "",
        urn.resource_type,
        "IF NOT EXISTS" if if_not_exists else "",
        urn.fqn,
        props.render(data),
    )


################ Update functions


def update_resource(urn: URN, data: dict, props: Props) -> str:
    return getattr(__this__, f"update_{urn.resource_label}", update__default)(urn, data, props)


def update__default(urn: URN, data: dict, props: Props) -> str:
    attr, new_value = data.popitem()
    attr = attr.lower()
    if new_value is None:
        return tidy_sql("ALTER", urn.resource_type, urn.fqn, "UNSET", attr)
    elif attr == "name":
        return tidy_sql("ALTER", urn.resource_type, urn.fqn, "RENAME TO", new_value)
    elif attr == "owner":
        raise NotImplementedError
    else:
        return tidy_sql(
            "ALTER",
            urn.resource_type,
            urn.fqn,
            "SET",
            props.render({attr: new_value}),
        )


def update_masking_policy(urn: URN, data: dict, props: Props) -> str:
    attr, new_value = data.popitem()
    attr = attr.lower()
    if attr == "body":
        return tidy_sql("ALTER", urn.resource_type, fqn_to_sql(urn.fqn), "SET BODY", props.render({"body": new_value}))
    else:
        return update__default(urn, {attr: new_value}, props)


def update_account_parameter(urn: URN, data: dict, props: Props) -> str:
    return create_account_parameter(urn, data, props)


def update_event_table(urn: URN, data: dict, props: Props) -> str:
    new_urn = URN(ResourceType.TABLE, urn.fqn, urn.account_locator)
    return update__default(new_urn, data, props)


def update_procedure(urn: URN, data: dict, props: Props) -> str:
    if "execute_as" in data:
        return tidy_sql(
            "ALTER",
            urn.resource_type,
            urn.fqn,
            "EXECUTE AS",
            data["execute_as"],
        )
    else:
        return update__default(urn, data, props)


def update_role_grant(urn: URN, data: dict, props: Props) -> str:
    raise NotImplementedError


def update_scanner_package(urn: URN, data: dict, props: Props) -> str:
    package_name = f"'{urn.fqn.name}'"
    attr, new_value = data.popitem()
    if attr == "schedule":
        new_value = f"'USING CRON {new_value}'"
    else:
        new_value = f"'{new_value}'"
    return tidy_sql(
        "CALL SNOWFLAKE.TRUST_CENTER.SET_CONFIGURATION(",
        f"'{attr}',",
        new_value,
        ",",
        package_name,
        ")",
    )


def update_schema(urn: URN, data: dict, props: Props) -> str:
    attr, new_value = data.popitem()
    attr = attr.lower()
    if new_value is None:
        return tidy_sql("ALTER SCHEMA", urn.fqn, "UNSET", attr)
    elif attr == "name":
        return tidy_sql("ALTER SCHEMA", urn.fqn, "RENAME TO", new_value)
    elif attr == "owner":
        raise NotImplementedError
    elif attr == "transient":
        raise Exception("Cannot change transient property of schema")
    elif attr == "managed_access":
        return tidy_sql("ALTER SCHEMA", urn.fqn, "ENABLE" if new_value else "DISABLE", "MANAGED ACCESS")
    else:
        new_value = f"'{new_value}'" if isinstance(new_value, str) else new_value
        return tidy_sql("ALTER SCHEMA", urn.fqn, "SET", attr, "=", new_value)


def update_table(urn: URN, data: dict, props: Props) -> str:
    attr, new_value = data.popitem()
    attr = attr.lower()
    if attr == "columns":
        raise NotImplementedError(data)
    else:
        return update__default(urn, {attr: new_value}, props)


# FIXME
# The AFTER attribute on tasks is one of the only attributes that is stateful,
# which means that you need to know the current value in order to modify it.
# This is a problem because we don't have a concept of "current value" for lifecycle updates
# and so we can't know what value to set.
def update_task(urn: URN, data: dict, props: Props) -> str:
    attr, new_value = data.popitem()
    attr = attr.lower()
    if attr == "as_":
        return tidy_sql("ALTER TASK", urn.fqn, "MODIFY", "AS", new_value)
    # elif attr == "after":
    #     if new_value is None:
    #         return tidy_sql("ALTER TASK", urn.fqn, "MODIFY", "AFTER", "NONE")
    #     else:
    #         return tidy_sql("ALTER TASK", urn.fqn, "MODIFY", "AFTER", ",".join([f"'{name}'" for name in new_value]))
    elif attr == "when":
        if new_value is None:
            return tidy_sql("ALTER TASK", urn.fqn, "REMOVE", "WHEN")
        else:
            return tidy_sql("ALTER TASK", urn.fqn, "MODIFY", "WHEN", new_value)
    elif attr == "state":
        change_verb = "RESUME" if new_value == "STARTED" else "SUSPEND"
        return tidy_sql("ALTER TASK", urn.fqn, change_verb)
    else:
        return update__default(urn, {attr: new_value}, props)


def update_iceberg_table(urn: URN, data: dict, props: Props) -> str:
    attr, new_value = data.popitem()
    attr = attr.lower()
    if attr == "columns":
        raise NotImplementedError(data)
    else:
        return update__default(urn, {attr: new_value}, props)


################ Drop functions


def drop_resource(urn: URN, data: dict, if_exists: bool = False) -> str:
    return getattr(__this__, f"drop_{urn.resource_label}", drop__default)(urn, data, if_exists=if_exists)


def drop__default(urn: URN, data: dict, if_exists: bool) -> str:
    return tidy_sql(
        "DROP",
        urn.resource_type,
        "IF EXISTS" if if_exists else "",
        fqn_to_sql(urn.fqn),
    )


def drop_account_parameter(urn: URN, data: dict, if_exists: bool) -> str:
    return tidy_sql(
        "ALTER",
        "ACCOUNT",
        "UNSET",
        urn.fqn.name,
    )


def drop_database(urn: URN, data: dict, if_exists: bool) -> str:
    return tidy_sql(
        "DROP",
        urn.resource_type,
        "IF EXISTS" if if_exists else "",
        urn.fqn,
        "RESTRICT",
    )


def drop_database_role_grant(urn: URN, data: dict, **kwargs):

    from_type = "ROLE" if data["to_role"] else "DATABASE ROLE"
    from_name = data["to_role"] if data["to_role"] else data["to_database_role"]

    return tidy_sql(
        "REVOKE DATABASE ROLE",
        ResourceName(data["database_role"]),
        "FROM",
        from_type,
        ResourceName(from_name),
    )


def drop_function(urn: URN, data: dict, if_exists: bool) -> str:
    return tidy_sql(
        "DROP",
        urn.resource_type,
        "IF EXISTS" if if_exists else "",
        urn.fqn,
    )


def drop_grant(urn: URN, data: dict, **kwargs):
    if data["priv"] == "OWNERSHIP":
        raise NotImplementedError
    if data["grant_type"] == GrantType.FUTURE:
        return tidy_sql(
            "REVOKE",
            data["priv"],
            "ON FUTURE",
            pluralize(data["items_type"]).upper(),
            "IN",
            data["on_type"],
            data["on"],
            "FROM",
            data["to"],
        )
    elif data["grant_type"] == GrantType.ALL:
        return tidy_sql(
            "REVOKE",
            data["priv"],
            "ON ALL",
            data["items_type"],
            "IN",
            data["on_type"],
            data["on"],
        )
    else:
        return tidy_sql(
            "REVOKE",
            data["priv"],
            "ON",
            data["on_type"],
            data["on"] if data["on_type"] != "ACCOUNT" else "",
            "FROM",
            data["to"],
            # "CASCADE" if cascade else "RESTRICT",
        )


def drop_procedure(urn: URN, data: dict, if_exists: bool) -> str:
    return tidy_sql(
        "DROP",
        urn.resource_type,
        "IF EXISTS" if if_exists else "",
        urn.fqn,
        # data["returns"],
    )


def drop_role_grant(urn: URN, data: dict, **kwargs):
    return tidy_sql(
        "REVOKE ROLE",
        ResourceName(data["role"]),
        "FROM",
        "ROLE" if data.get("to_role") else "USER",
        ResourceName(data["to_role"] if data.get("to_role") else data["to_user"]),
    )


def drop_scanner_package(urn: URN, data: dict, **kwargs) -> str:
    package_name = f"'{urn.fqn.name}'"
    return tidy_sql(
        "CALL SNOWFLAKE.TRUST_CENTER.SET_CONFIGURATION(",
        "'ENABLED',",
        "'FALSE',",
        package_name,
        ")",
    )


def drop_tag_masking_policy_reference(urn: URN, data: dict, **kwargs) -> str:
    return tidy_sql(
        "ALTER TAG",
        data["tag_name"],
        "UNSET MASKING POLICY",
        data["masking_policy_name"],
    )


################ Transfer functions


def transfer_resource(
    urn: URN,
    owner: str,
    owner_resource_type: ResourceType,
    copy_current_grants: bool = False,
    revoke_current_grants: bool = False,
) -> str:
    return getattr(__this__, f"transfer_{urn.resource_label}", transfer__default)(
        urn,
        owner,
        owner_resource_type,
        copy_current_grants,
        revoke_current_grants,
    )


def transfer__default(
    urn: URN,
    owner: str,
    owner_resource_type: ResourceType,
    copy_current_grants: bool = False,
    revoke_current_grants: bool = False,
) -> str:
    return tidy_sql(
        "GRANT OWNERSHIP ON",
        urn.resource_type,
        urn.fqn,
        "TO",
        owner_resource_type,
        owner,
        "REVOKE CURRENT GRANTS" if revoke_current_grants else "",
        "COPY CURRENT GRANTS" if copy_current_grants else "",
    )



"""
User-friendly error message formatting for Snowcap exceptions.
"""

import difflib
from typing import Optional

from .identifiers import URN


def _get_resource_display_name(urn: URN) -> str:
    """
    Get a display name for a resource that includes full qualification where relevant.

    For schema-scoped resources, includes database.schema.name.
    For database-scoped resources, includes database.name.
    For functions/procedures, includes arg types.
    """
    fqn = urn.fqn
    parts = []

    if fqn.database:
        parts.append(str(fqn.database))
    if fqn.schema:
        parts.append(str(fqn.schema))
    parts.append(str(fqn.name))

    name = ".".join(parts)

    # Add arg types for functions/procedures
    if fqn.arg_types is not None:
        name += f"({', '.join(fqn.arg_types)})"

    return name


def format_missing_resource_error(
    missing_urn: URN,
    required_by_urn: Optional[URN] = None,
    available_names: Optional[list[str]] = None,
) -> str:
    """
    Format a user-friendly error message for missing resources.

    Args:
        missing_urn: The URN of the resource that was not found
        required_by_urn: The URN of the resource that references the missing resource
        available_names: List of available resource names to suggest alternatives

    Returns:
        A formatted error message with context and suggestions
    """
    # Extract readable names
    resource_type = missing_urn.resource_label.replace("_", " ").title()
    resource_name = _get_resource_display_name(missing_urn)

    # Build message
    msg = f'{resource_type} "{resource_name}" not found.'

    # Add context about what references it
    if required_by_urn:
        msg += f"\n  Referenced by: {_format_reference(required_by_urn)}"

    # Add suggestions if available
    if available_names:
        # For matching, use just the base name for better suggestions
        base_name = str(missing_urn.fqn.name)
        suggestions = difflib.get_close_matches(base_name, available_names, n=3, cutoff=0.4)
        if suggestions:
            msg += f"\n  Did you mean: {', '.join(suggestions)}?"

    return msg


def _format_reference(urn: URN) -> str:
    """
    Format a URN as a human-readable reference.

    Args:
        urn: The URN to format

    Returns:
        A human-readable description of the reference
    """
    resource_type = urn.resource_label.replace("_", " ")
    params = urn.fqn.params

    # Special handling for grants
    if "grant" in resource_type:
        if "user" in params:
            return f'{resource_type} to user "{params["user"]}"'
        elif "role" in params:
            return f'{resource_type} to role "{params["role"]}"'
        # Fallback for other grant types - include what we know
        grant_info = []
        if "on" in params:
            grant_info.append(f'on "{params["on"]}"')
        if "to" in params:
            grant_info.append(f'to "{params["to"]}"')
        if grant_info:
            return f'{resource_type} {" ".join(grant_info)}'

    return f'{resource_type} "{_get_resource_display_name(urn)}"'


def format_missing_container_error(container_urn: URN) -> str:
    """
    Format a user-friendly error message for missing container resources.

    Args:
        container_urn: The URN of the container that was not found

    Returns:
        A formatted error message
    """
    resource_type = container_urn.resource_label.replace("_", " ").title()
    resource_name = _get_resource_display_name(container_urn)

    return (
        f'{resource_type} "{resource_name}" not found.\n'
        f"  This resource is referenced but doesn't exist or isn't visible in the current session."
    )


def format_missing_pointer_error(
    urn: URN,
    available_names: Optional[list[str]] = None,
) -> str:
    """
    Format a user-friendly error message for missing resource pointers.

    Args:
        urn: The URN of the missing resource
        available_names: List of available resource names to suggest alternatives

    Returns:
        A formatted error message with suggestions
    """
    resource_type = urn.resource_label.replace("_", " ").title()
    resource_name = _get_resource_display_name(urn)

    msg = f'{resource_type} "{resource_name}" not found.'

    # Add suggestions if available
    if available_names:
        # For matching, use just the base name for better suggestions
        base_name = str(urn.fqn.name)
        suggestions = difflib.get_close_matches(base_name, available_names, n=3, cutoff=0.4)
        if suggestions:
            msg += f"\n  Did you mean: {', '.join(suggestions)}?"

    return msg


def format_invalid_key_error(
    invalid_keys: list[str],
    valid_keys: list[str],
    resource_type: str,
    resource_name: Optional[str] = None,
) -> tuple[str, dict[str, str]]:
    """
    Format a user-friendly error message for invalid keys in resources.

    Args:
        invalid_keys: List of invalid key names that were provided
        valid_keys: List of valid key names for this resource type
        resource_type: The type of resource (e.g., "Database", "Role")
        resource_name: Optional name of the specific resource instance

    Returns:
        A tuple of (formatted error message, dict of suggestions mapping invalid->valid keys)
    """
    suggestions = {}
    for key in invalid_keys:
        matches = difflib.get_close_matches(key, valid_keys, n=1, cutoff=0.6)
        if matches:
            suggestions[key] = matches[0]

    if len(invalid_keys) == 1:
        key = invalid_keys[0]
        if resource_name:
            msg = f'Invalid key "{key}" in {resource_type} "{resource_name}".'
        else:
            msg = f'Invalid key "{key}" in {resource_type}.'
        if key in suggestions:
            msg += f"\n  Did you mean: {suggestions[key]}?"
    else:
        keys_str = ", ".join(f'"{k}"' for k in invalid_keys)
        if resource_name:
            msg = f'Invalid keys {keys_str} in {resource_type} "{resource_name}".'
        else:
            msg = f'Invalid keys {keys_str} in {resource_type}.'
        for key, suggestion in suggestions.items():
            msg += f'\n  "{key}" -> Did you mean: {suggestion}?'

    return msg, suggestions


def format_invalid_role_grant_keys(
    invalid_keys: set[str],
    valid_keys: set[str],
) -> str:
    """
    Format a user-friendly error message for invalid keys in role_grants config.

    Args:
        invalid_keys: Set of invalid key names that were provided
        valid_keys: Set of valid key names for role_grants

    Returns:
        A formatted error message with suggestions
    """
    suggestions = {}
    for key in invalid_keys:
        matches = difflib.get_close_matches(key, list(valid_keys), n=1, cutoff=0.6)
        if matches:
            suggestions[key] = matches[0]

    if len(invalid_keys) == 1:
        key = list(invalid_keys)[0]
        msg = f'Invalid key "{key}" in role_grant.'
        if key in suggestions:
            msg += f"\n  Did you mean: {suggestions[key]}?"
    else:
        keys_str = ", ".join(f'"{k}"' for k in sorted(invalid_keys))
        msg = f"Invalid keys {keys_str} in role_grant."
        for key in sorted(invalid_keys):
            if key in suggestions:
                msg += f'\n  "{key}" -> Did you mean: {suggestions[key]}?'

    return msg



from typing import TYPE_CHECKING, Union

from .var import VarString

if TYPE_CHECKING:
    from snowcap.resources.role import DatabaseRole, Role

RoleRef = Union["Role", "DatabaseRole", VarString, str]



"""
Resource-specific unit tests for snowcap resources.

Each resource type has tests for:
1. Instantiation with minimal required properties
2. Instantiation with all properties
3. SQL generation (create_sql)
4. FQN and URN generation
5. Resource-specific validation rules
6. Property defaults
"""

import pytest
from unittest.mock import MagicMock

from snowcap import resources as res
from snowcap.enums import ResourceType, WarehouseSize
from snowcap.identifiers import FQN, URN
from snowcap.resource_name import ResourceName


class TestDatabase:
    """Tests for Database resource."""

    def test_database_minimal(self):
        """Test Database with minimal required properties."""
        db = res.Database(name="test_db")
        assert db.name == "test_db"
        assert db.resource_type == ResourceType.DATABASE

    def test_database_all_properties(self):
        """Test Database with all properties."""
        db = res.Database(
            name="test_db",
            transient=True,
            owner="ACCOUNTADMIN",
            data_retention_time_in_days=7,
            max_data_extension_time_in_days=28,
            default_ddl_collation="en_US",
            comment="Test database",
        )
        assert db._data.transient is True
        assert db._data.data_retention_time_in_days == 7
        assert db._data.max_data_extension_time_in_days == 28
        assert db._data.default_ddl_collation == "en_US"
        assert db._data.comment == "Test database"

    def test_database_fqn(self):
        """Test Database FQN generation."""
        db = res.Database(name="test_db")
        fqn = db.fqn
        assert fqn.name == "TEST_DB"
        assert fqn.database is None
        assert fqn.schema is None

    def test_database_public_schema_created(self):
        """Test that Database automatically creates PUBLIC schema."""
        db = res.Database(name="test_db")
        assert db.public_schema is not None
        assert db.public_schema.name == "PUBLIC"

    def test_database_snowflake_no_public_schema(self):
        """Test that SNOWFLAKE database does not create PUBLIC schema."""
        db = res.Database(name="SNOWFLAKE")
        assert not hasattr(db, "_public_schema") or db._public_schema is None

    def test_database_defaults(self):
        """Test Database property defaults."""
        db = res.Database(name="test_db")
        assert db._data.transient is False
        # data_retention and max_data_extension default to None (inherit from Snowflake defaults)
        assert db._data.data_retention_time_in_days is None
        assert db._data.max_data_extension_time_in_days is None

    def test_database_to_dict(self):
        """Test Database serialization."""
        db = res.Database(name="test_db", comment="Test")
        data = db.to_dict()
        assert data["name"] == "TEST_DB"
        assert data["comment"] == "Test"


class TestSchema:
    """Tests for Schema resource."""

    def test_schema_minimal(self):
        """Test Schema with minimal required properties."""
        schema = res.Schema(name="test_schema", database="test_db")
        assert schema.name == "test_schema"
        assert schema.resource_type == ResourceType.SCHEMA

    def test_schema_with_database_object(self):
        """Test Schema with Database object."""
        db = res.Database(name="test_db")
        schema = res.Schema(name="test_schema", database=db)
        assert schema.container is not None
        assert schema.container.name == "test_db"

    def test_schema_all_properties(self):
        """Test Schema with all properties."""
        schema = res.Schema(
            name="test_schema",
            database="test_db",
            transient=True,
            owner="SYSADMIN",
            data_retention_time_in_days=7,
            max_data_extension_time_in_days=28,
            default_ddl_collation="en_US",
            comment="Test schema",
        )
        assert schema._data.transient is True
        assert schema._data.data_retention_time_in_days == 7

    def test_schema_fqn(self):
        """Test Schema FQN generation."""
        schema = res.Schema(name="test_schema", database="test_db")
        fqn = schema.fqn
        assert fqn.name == "TEST_SCHEMA"
        assert str(fqn.database) == "TEST_DB"

    def test_schema_from_fqn_string(self):
        """Test Schema creation from fully qualified name string."""
        schema = res.Schema(name="test_db.test_schema")
        assert schema.name == "test_schema"
        assert schema.container.name == "test_db"

    def test_schema_defaults(self):
        """Test Schema property defaults."""
        schema = res.Schema(name="test_schema", database="test_db")
        assert schema._data.transient is False
        # data_retention defaults to None (inherit from database)
        assert schema._data.data_retention_time_in_days is None


class TestRole:
    """Tests for Role resource."""

    def test_role_minimal(self):
        """Test Role with minimal required properties."""
        role = res.Role(name="test_role")
        assert role.name == "test_role"
        assert role.resource_type == ResourceType.ROLE

    def test_role_all_properties(self):
        """Test Role with all properties."""
        role = res.Role(
            name="test_role",
            owner="USERADMIN",
            comment="Test role",
        )
        assert role._data.comment == "Test role"

    def test_role_fqn(self):
        """Test Role FQN generation."""
        role = res.Role(name="test_role")
        fqn = role.fqn
        assert fqn.name == "TEST_ROLE"

    def test_role_defaults(self):
        """Test Role property defaults."""
        role = res.Role(name="test_role")
        assert role._data.owner.name == "USERADMIN"

    def test_role_to_dict(self):
        """Test Role serialization."""
        role = res.Role(name="test_role", comment="Test")
        data = role.to_dict()
        assert data["name"] == "TEST_ROLE"
        assert data["comment"] == "Test"


class TestDatabaseRole:
    """Tests for DatabaseRole resource."""

    def test_database_role_minimal(self):
        """Test DatabaseRole with minimal required properties."""
        db_role = res.DatabaseRole(name="test_db_role", database="test_db")
        assert db_role.name == "test_db_role"
        assert db_role.resource_type == ResourceType.DATABASE_ROLE

    def test_database_role_all_properties(self):
        """Test DatabaseRole with all properties."""
        db_role = res.DatabaseRole(
            name="test_db_role",
            database="test_db",
            owner="USERADMIN",
            comment="Test database role",
        )
        assert db_role._data.comment == "Test database role"

    def test_database_role_fqn(self):
        """Test DatabaseRole FQN generation."""
        db_role = res.DatabaseRole(name="test_db_role", database="test_db")
        fqn = db_role.fqn
        assert fqn.name == "TEST_DB_ROLE"
        assert str(fqn.database) == "TEST_DB"

    def test_database_role_database_property(self):
        """Test DatabaseRole database property."""
        db_role = res.DatabaseRole(name="test_db_role", database="test_db")
        assert db_role.database == "test_db"


class TestWarehouse:
    """Tests for Warehouse resource."""

    def test_warehouse_minimal(self):
        """Test Warehouse with minimal required properties."""
        wh = res.Warehouse(name="test_wh")
        assert wh.name == "test_wh"
        assert wh.resource_type == ResourceType.WAREHOUSE

    def test_warehouse_all_properties(self):
        """Test Warehouse with all properties."""
        wh = res.Warehouse(
            name="test_wh",
            owner="SYSADMIN",
            warehouse_type="STANDARD",
            warehouse_size="MEDIUM",
            max_cluster_count=5,
            min_cluster_count=1,
            scaling_policy="ECONOMY",
            auto_suspend=300,
            auto_resume=True,
            initially_suspended=True,
            comment="Test warehouse",
            enable_query_acceleration=True,
            query_acceleration_max_scale_factor=4,
            max_concurrency_level=16,
            statement_queued_timeout_in_seconds=30,
            statement_timeout_in_seconds=3600,
        )
        assert wh._data.warehouse_size == WarehouseSize.MEDIUM
        assert wh._data.max_cluster_count == 5
        assert wh._data.auto_suspend == 300
        assert wh._data.initially_suspended is True

    def test_warehouse_fqn(self):
        """Test Warehouse FQN generation."""
        wh = res.Warehouse(name="test_wh")
        fqn = wh.fqn
        assert fqn.name == "TEST_WH"

    def test_warehouse_defaults(self):
        """Test Warehouse property defaults."""
        wh = res.Warehouse(name="test_wh")
        assert wh._data.warehouse_size == WarehouseSize.XSMALL
        assert wh._data.auto_suspend == 600
        assert wh._data.auto_resume is True
        assert wh._data.max_cluster_count == 1

    def test_warehouse_size_enum_conversion(self):
        """Test Warehouse size enum conversion from string."""
        wh = res.Warehouse(name="test_wh", warehouse_size="LARGE")
        assert wh._data.warehouse_size == WarehouseSize.LARGE


class TestUser:
    """Tests for User resource."""

    def test_user_minimal(self):
        """Test User with minimal required properties."""
        user = res.User(name="test_user")
        assert user.name == "test_user"
        assert user.resource_type == ResourceType.USER

    def test_user_all_properties(self):
        """Test User with all properties."""
        user = res.User(
            name="test_user",
            login_name="testuser",
            display_name="Test User",
            first_name="Test",
            last_name="User",
            email="test@example.com",
            disabled=False,
            must_change_password=True,
            default_warehouse="test_wh",
            default_role="test_role",
            comment="Test user",
        )
        assert user._data.login_name == "TESTUSER"
        assert user._data.display_name == "Test User"
        assert user._data.email == "test@example.com"

    def test_user_fqn(self):
        """Test User FQN generation."""
        user = res.User(name="test_user")
        fqn = user.fqn
        assert fqn.name == "TEST_USER"


class TestTable:
    """Tests for Table resource."""

    def test_table_minimal(self):
        """Test Table with minimal required properties."""
        table = res.Table(
            name="test_table",
            database="test_db",
            schema="test_schema",
            columns=[{"name": "id", "data_type": "INT"}],
        )
        assert table.name == "test_table"
        assert table.resource_type == ResourceType.TABLE

    def test_table_all_properties(self):
        """Test Table with all properties."""
        table = res.Table(
            name="test_table",
            database="test_db",
            schema="test_schema",
            columns=[
                {"name": "id", "data_type": "INT", "not_null": True},
                {"name": "name", "data_type": "VARCHAR(100)"},
            ],
            cluster_by=["id"],
            transient=True,
            comment="Test table",
        )
        assert len(table._data.columns) == 2
        assert table._data.transient is True

    def test_table_fqn(self):
        """Test Table FQN generation."""
        table = res.Table(
            name="test_table",
            database="test_db",
            schema="test_schema",
            columns=[{"name": "id", "data_type": "INT"}],
        )
        fqn = table.fqn
        assert fqn.name == "TEST_TABLE"
        assert str(fqn.database) == "TEST_DB"
        assert str(fqn.schema) == "TEST_SCHEMA"

    def test_table_from_fqn_string(self):
        """Test Table creation from fully qualified name string."""
        table = res.Table(
            name="test_db.test_schema.test_table",
            columns=[{"name": "id", "data_type": "INT"}],
        )
        assert table.name == "test_table"
        assert table.container.name == "test_schema"
        assert table.container.container.name == "test_db"


class TestView:
    """Tests for View resource."""

    def test_view_minimal(self):
        """Test View with minimal required properties."""
        view = res.View(
            name="test_view",
            database="test_db",
            schema="test_schema",
            as_="SELECT 1 AS col",
        )
        assert view.name == "test_view"
        assert view.resource_type == ResourceType.VIEW

    def test_view_with_columns(self):
        """Test View with columns."""
        view = res.View(
            name="test_view",
            database="test_db",
            schema="test_schema",
            columns=[{"name": "col1"}],
            as_="SELECT 1 AS col1",
        )
        assert len(view._data.columns) == 1

    def test_view_fqn(self):
        """Test View FQN generation."""
        view = res.View(
            name="test_view",
            database="test_db",
            schema="test_schema",
            as_="SELECT 1",
        )
        fqn = view.fqn
        assert fqn.name == "TEST_VIEW"
        assert str(fqn.database) == "TEST_DB"
        assert str(fqn.schema) == "TEST_SCHEMA"

    def test_view_empty_columns_rejected(self):
        """Test View with empty columns list is rejected."""
        with pytest.raises(ValueError):
            res.View(
                name="test_view",
                database="test_db",
                schema="test_schema",
                columns=[],
                as_="SELECT 1",
            )


class TestGrant:
    """Tests for Grant resource."""

    def test_grant_minimal(self):
        """Test Grant with minimal required properties."""
        grant = res.Grant(
            priv="USAGE",
            on_database="test_db",
            to="test_role",
        )
        assert grant.resource_type == ResourceType.GRANT

    def test_grant_on_schema(self):
        """Test Grant on schema."""
        grant = res.Grant(
            priv="USAGE",
            on_schema="test_db.test_schema",
            to="test_role",
        )
        assert grant._data.on_type == "SCHEMA"

    def test_grant_on_table(self):
        """Test Grant on table."""
        grant = res.Grant(
            priv="SELECT",
            on_table="test_db.test_schema.test_table",
            to="test_role",
        )
        assert grant._data.on_type == "TABLE"

    def test_grant_with_grant_option(self):
        """Test Grant with grant option."""
        grant = res.Grant(
            priv="USAGE",
            on_database="test_db",
            to="test_role",
            grant_option=True,
        )
        assert grant._data.grant_option is True


class TestRoleGrant:
    """Tests for RoleGrant resource."""

    def test_role_grant_to_role(self):
        """Test RoleGrant to another role."""
        grant = res.RoleGrant(
            role="child_role",
            to_role="parent_role",
        )
        assert grant.resource_type == ResourceType.ROLE_GRANT

    def test_role_grant_to_user(self):
        """Test RoleGrant to user."""
        grant = res.RoleGrant(
            role="test_role",
            to_user="test_user",
        )
        assert grant._data.to_user.name == "test_user"


class TestTask:
    """Tests for Task resource."""

    def test_task_minimal(self):
        """Test Task with minimal required properties."""
        task = res.Task(
            name="test_task",
            database="test_db",
            schema="test_schema",
            as_="SELECT 1",
        )
        assert task.name == "test_task"
        assert task.resource_type == ResourceType.TASK

    def test_task_with_schedule(self):
        """Test Task with schedule."""
        task = res.Task(
            name="test_task",
            database="test_db",
            schema="test_schema",
            schedule="1 MINUTE",
            as_="SELECT 1",
        )
        assert task._data.schedule == "1 MINUTE"

    def test_task_with_warehouse(self):
        """Test Task with warehouse."""
        task = res.Task(
            name="test_task",
            database="test_db",
            schema="test_schema",
            warehouse="test_wh",
            as_="SELECT 1",
        )
        assert task._data.warehouse.name == "test_wh"

    def test_task_fqn(self):
        """Test Task FQN generation."""
        task = res.Task(
            name="test_task",
            database="test_db",
            schema="test_schema",
            as_="SELECT 1",
        )
        fqn = task.fqn
        assert fqn.name == "TEST_TASK"


class TestSequence:
    """Tests for Sequence resource."""

    def test_sequence_minimal(self):
        """Test Sequence with minimal required properties."""
        seq = res.Sequence(
            name="test_seq",
            database="test_db",
            schema="test_schema",
        )
        assert seq.name == "test_seq"
        assert seq.resource_type == ResourceType.SEQUENCE

    def test_sequence_all_properties(self):
        """Test Sequence with all properties."""
        seq = res.Sequence(
            name="test_seq",
            database="test_db",
            schema="test_schema",
            start=100,
            increment=10,
            comment="Test sequence",
        )
        assert seq._data.start == 100
        assert seq._data.increment == 10


class TestInternalStage:
    """Tests for InternalStage resource."""

    def test_internal_stage_minimal(self):
        """Test InternalStage with minimal required properties."""
        stage = res.InternalStage(
            name="test_stage",
            database="test_db",
            schema="test_schema",
        )
        assert stage.name == "test_stage"
        assert stage.resource_type == ResourceType.STAGE

    def test_internal_stage_all_properties(self):
        """Test InternalStage with all properties."""
        stage = res.InternalStage(
            name="test_stage",
            database="test_db",
            schema="test_schema",
            directory={"enable": True},
            comment="Test stage",
        )
        assert stage._data.directory is not None


class TestNetworkPolicy:
    """Tests for NetworkPolicy resource."""

    def test_network_policy_minimal(self):
        """Test NetworkPolicy with minimal required properties."""
        policy = res.NetworkPolicy(
            name="test_policy",
        )
        assert policy.name == "test_policy"
        assert policy.resource_type == ResourceType.NETWORK_POLICY

    def test_network_policy_all_properties(self):
        """Test NetworkPolicy with all properties."""
        policy = res.NetworkPolicy(
            name="test_policy",
            allowed_ip_list=["192.168.1.0/24"],
            blocked_ip_list=["10.0.0.0/8"],
            comment="Test policy",
        )
        assert policy._data.allowed_ip_list == ["192.168.1.0/24"]
        assert policy._data.blocked_ip_list == ["10.0.0.0/8"]


class TestNetworkRule:
    """Tests for NetworkRule resource."""

    def test_network_rule_minimal(self):
        """Test NetworkRule with minimal required properties."""
        rule = res.NetworkRule(
            name="test_rule",
            database="test_db",
            schema="test_schema",
            type="IPV4",
            value_list=["192.168.1.0/24"],
            mode="INGRESS",
        )
        assert rule.name == "test_rule"
        assert rule.resource_type == ResourceType.NETWORK_RULE


class TestPipe:
    """Tests for Pipe resource."""

    def test_pipe_minimal(self):
        """Test Pipe with minimal required properties."""
        pipe = res.Pipe(
            name="test_pipe",
            database="test_db",
            schema="test_schema",
            as_="COPY INTO t FROM @s",
        )
        assert pipe.name == "test_pipe"
        assert pipe.resource_type == ResourceType.PIPE

    def test_pipe_all_properties(self):
        """Test Pipe with all properties."""
        pipe = res.Pipe(
            name="test_pipe",
            database="test_db",
            schema="test_schema",
            as_="COPY INTO t FROM @s",
            auto_ingest=True,
            comment="Test pipe",
        )
        assert pipe._data.auto_ingest is True


class TestAlert:
    """Tests for Alert resource."""

    def test_alert_minimal(self):
        """Test Alert with minimal required properties."""
        alert = res.Alert(
            name="test_alert",
            database="test_db",
            schema="test_schema",
            warehouse="test_wh",
            schedule="1 MINUTE",
            condition="SELECT 1",
            then="SELECT 1",  # 'then' is the action parameter
        )
        assert alert.name == "test_alert"
        assert alert.resource_type == ResourceType.ALERT


class TestResourceMonitor:
    """Tests for ResourceMonitor resource."""

    def test_resource_monitor_minimal(self):
        """Test ResourceMonitor with minimal required properties."""
        rm = res.ResourceMonitor(
            name="test_rm",
        )
        assert rm.name == "test_rm"
        assert rm.resource_type == ResourceType.RESOURCE_MONITOR

    def test_resource_monitor_all_properties(self):
        """Test ResourceMonitor with all properties."""
        rm = res.ResourceMonitor(
            name="test_rm",
            credit_quota=1000,
            frequency="MONTHLY",
            start_timestamp="2024-01-01 00:00",
        )
        assert rm._data.credit_quota == 1000


class TestTag:
    """Tests for Tag resource."""

    def test_tag_minimal(self):
        """Test Tag with minimal required properties."""
        tag = res.Tag(
            name="test_tag",
            database="test_db",
            schema="test_schema",
        )
        assert tag.name == "test_tag"
        assert tag.resource_type == ResourceType.TAG

    def test_tag_with_allowed_values(self):
        """Test Tag with allowed values."""
        tag = res.Tag(
            name="test_tag",
            database="test_db",
            schema="test_schema",
            allowed_values=["value1", "value2"],
        )
        assert tag._data.allowed_values == ["value1", "value2"]


class TestStream:
    """Tests for Stream resources."""

    def test_table_stream_minimal(self):
        """Test TableStream with minimal required properties."""
        stream = res.TableStream(
            name="test_stream",
            database="test_db",
            schema="test_schema",
            on_table="source_table",
        )
        assert stream.name == "test_stream"
        assert stream.resource_type == ResourceType.STREAM

    def test_view_stream_minimal(self):
        """Test ViewStream with minimal required properties."""
        stream = res.ViewStream(
            name="test_stream",
            database="test_db",
            schema="test_schema",
            on_view="source_view",
        )
        assert stream.name == "test_stream"

    def test_stage_stream_minimal(self):
        """Test StageStream with minimal required properties."""
        stream = res.StageStream(
            name="test_stream",
            database="test_db",
            schema="test_schema",
            on_stage="source_stage",
        )
        assert stream.name == "test_stream"


class TestFileFormat:
    """Tests for FileFormat resources."""

    def test_csv_file_format_minimal(self):
        """Test CSVFileFormat with minimal required properties."""
        ff = res.CSVFileFormat(
            name="test_ff",
            database="test_db",
            schema="test_schema",
        )
        assert ff.name == "test_ff"
        assert ff.resource_type == ResourceType.FILE_FORMAT

    def test_csv_file_format_all_properties(self):
        """Test CSVFileFormat with all properties."""
        ff = res.CSVFileFormat(
            name="test_ff",
            database="test_db",
            schema="test_schema",
            field_delimiter="|",
            record_delimiter="\n",
            skip_header=1,
        )
        assert ff._data.field_delimiter == "|"

    def test_json_file_format_minimal(self):
        """Test JSONFileFormat with minimal required properties."""
        ff = res.JSONFileFormat(
            name="test_ff",
            database="test_db",
            schema="test_schema",
        )
        assert ff.name == "test_ff"

    def test_parquet_file_format_minimal(self):
        """Test ParquetFileFormat with minimal required properties."""
        ff = res.ParquetFileFormat(
            name="test_ff",
            database="test_db",
            schema="test_schema",
        )
        assert ff.name == "test_ff"


class TestStorageIntegration:
    """Tests for StorageIntegration resources."""

    def test_s3_storage_integration_minimal(self):
        """Test S3StorageIntegration with minimal required properties."""
        si = res.S3StorageIntegration(
            name="test_si",
            enabled=True,
            storage_aws_role_arn="arn:aws:iam::123456789012:role/MyRole",
            storage_allowed_locations=["s3://bucket/path/"],
        )
        assert si.name == "test_si"
        assert si.resource_type == ResourceType.STORAGE_INTEGRATION

    def test_gcs_storage_integration_minimal(self):
        """Test GCSStorageIntegration with minimal required properties."""
        si = res.GCSStorageIntegration(
            name="test_si",
            enabled=True,
            storage_allowed_locations=["gcs://bucket/path/"],
        )
        assert si.name == "test_si"

    def test_azure_storage_integration_minimal(self):
        """Test AzureStorageIntegration with minimal required properties."""
        si = res.AzureStorageIntegration(
            name="test_si",
            enabled=True,
            azure_tenant_id="tenant-id",
            storage_allowed_locations=["azure://container/path/"],
        )
        assert si.name == "test_si"


class TestExternalAccessIntegration:
    """Tests for ExternalAccessIntegration resource."""

    def test_external_access_integration_minimal(self):
        """Test ExternalAccessIntegration with minimal required properties."""
        # allowed_network_rules must have at least one element
        eai = res.ExternalAccessIntegration(
            name="test_eai",
            allowed_network_rules=["db.schema.rule1"],
        )
        assert eai.name == "test_eai"
        assert eai.resource_type == ResourceType.EXTERNAL_ACCESS_INTEGRATION

    def test_external_access_integration_all_properties(self):
        """Test ExternalAccessIntegration with all properties."""
        eai = res.ExternalAccessIntegration(
            name="test_eai",
            allowed_network_rules=["db.schema.rule1"],
            allowed_api_authentication_integrations=["auth_int1"],
            enabled=True,
            comment="Test EAI",
        )
        assert len(eai._data.allowed_network_rules) == 1


class TestSecret:
    """Tests for Secret resources."""

    def test_password_secret_minimal(self):
        """Test PasswordSecret with minimal required properties."""
        secret = res.PasswordSecret(
            name="test_secret",
            database="test_db",
            schema="test_schema",
            username="user",
            password="pass",
        )
        assert secret.name == "test_secret"
        assert secret.resource_type == ResourceType.SECRET

    def test_generic_secret_minimal(self):
        """Test GenericSecret with minimal required properties."""
        secret = res.GenericSecret(
            name="test_secret",
            database="test_db",
            schema="test_schema",
            secret_string="secret_value",
        )
        assert secret.name == "test_secret"


class TestPasswordPolicy:
    """Tests for PasswordPolicy resource."""

    def test_password_policy_minimal(self):
        """Test PasswordPolicy with minimal required properties."""
        policy = res.PasswordPolicy(
            name="test_policy",
            database="test_db",
            schema="test_schema",
        )
        assert policy.name == "test_policy"
        assert policy.resource_type == ResourceType.PASSWORD_POLICY

    def test_password_policy_all_properties(self):
        """Test PasswordPolicy with all properties."""
        policy = res.PasswordPolicy(
            name="test_policy",
            database="test_db",
            schema="test_schema",
            password_min_length=12,
            password_max_length=256,
            password_min_upper_case_chars=2,
            password_min_lower_case_chars=2,
            password_min_numeric_chars=2,
            password_min_special_chars=2,
        )
        assert policy._data.password_min_length == 12


class TestProcedure:
    """Tests for Procedure resources."""

    def test_python_stored_procedure_minimal(self):
        """Test PythonStoredProcedure with minimal required properties."""
        proc = res.PythonStoredProcedure(
            name="test_proc()",
            database="test_db",
            schema="test_schema",
            args=[],  # Required parameter
            returns="VARCHAR",
            runtime_version="3.8",
            handler="handler",
            packages=["snowflake-snowpark-python"],
            as_="def handler(): return 'hello'",
        )
        # Function names strip the parentheses
        assert "test_proc" in str(proc.name).lower()
        assert proc.resource_type == ResourceType.PROCEDURE


class TestFunction:
    """Tests for Function resources."""

    def test_python_udf_minimal(self):
        """Test PythonUDF with minimal required properties."""
        func = res.PythonUDF(
            name="test_func()",
            database="test_db",
            schema="test_schema",
            args=[],  # Required parameter
            returns="VARCHAR",
            runtime_version="3.8",
            handler="handler",
            as_="def handler(): return 'hello'",
        )
        # Function names strip the parentheses
        assert "test_func" in str(func.name).lower()
        assert func.resource_type == ResourceType.FUNCTION

    def test_javascript_udf_minimal(self):
        """Test JavascriptUDF with minimal required properties."""
        func = res.JavascriptUDF(
            name="test_func()",
            database="test_db",
            schema="test_schema",
            args=[],  # Required parameter
            returns="VARCHAR",
            as_="return 'hello';",
        )
        # Function names strip the parentheses
        assert "test_func" in str(func.name).lower()


class TestDynamicTable:
    """Tests for DynamicTable resource."""

    def test_dynamic_table_minimal(self):
        """Test DynamicTable with minimal required properties."""
        dt = res.DynamicTable(
            name="test_dt",
            database="test_db",
            schema="test_schema",
            columns=[{"name": "col1"}],  # Required parameter
            target_lag="1 HOUR",
            warehouse="test_wh",
            as_="SELECT * FROM source",
        )
        assert dt.name == "test_dt"
        assert dt.resource_type == ResourceType.DYNAMIC_TABLE


class TestMaterializedView:
    """Tests for MaterializedView resource."""

    def test_materialized_view_minimal(self):
        """Test MaterializedView with minimal required properties."""
        mv = res.MaterializedView(
            name="test_mv",
            database="test_db",
            schema="test_schema",
            as_="SELECT * FROM source",
        )
        assert mv.name == "test_mv"
        assert mv.resource_type == ResourceType.MATERIALIZED_VIEW


class TestAPIIntegration:
    """Tests for APIIntegration resource."""

    def test_api_integration_minimal(self):
        """Test APIIntegration with minimal required properties."""
        api = res.APIIntegration(
            name="test_api",
            api_provider="AWS_API_GATEWAY",
            api_aws_role_arn="arn:aws:iam::123456789012:role/MyRole",
            api_allowed_prefixes=["https://api.example.com/"],
            enabled=True,
        )
        assert api.name == "test_api"
        assert api.resource_type == ResourceType.API_INTEGRATION


class TestShare:
    """Tests for Share resource."""

    def test_share_minimal(self):
        """Test Share with minimal required properties."""
        share = res.Share(name="test_share")
        assert share.name == "test_share"
        assert share.resource_type == ResourceType.SHARE


class TestResourceCommon:
    """Common tests that apply to all resources."""

    @pytest.mark.parametrize(
        "resource_cls,kwargs",
        [
            (res.Database, {"name": "test"}),
            (res.Schema, {"name": "test", "database": "db"}),
            (res.Role, {"name": "test"}),
            (res.Warehouse, {"name": "test"}),
            (res.User, {"name": "test"}),
            (res.Table, {"name": "test", "database": "db", "schema": "sch", "columns": [{"name": "id", "data_type": "INT"}]}),
            (res.View, {"name": "test", "database": "db", "schema": "sch", "as_": "SELECT 1"}),
            (res.NetworkPolicy, {"name": "test"}),
        ],
    )
    def test_resource_has_name(self, resource_cls, kwargs):
        """Test that all resources have a name property."""
        resource = resource_cls(**kwargs)
        assert resource.name is not None

    @pytest.mark.parametrize(
        "resource_cls,kwargs",
        [
            (res.Database, {"name": "test"}),
            (res.Schema, {"name": "test", "database": "db"}),
            (res.Role, {"name": "test"}),
            (res.Warehouse, {"name": "test"}),
        ],
    )
    def test_resource_has_fqn(self, resource_cls, kwargs):
        """Test that all named resources have a FQN property."""
        resource = resource_cls(**kwargs)
        assert resource.fqn is not None
        assert isinstance(resource.fqn, FQN)

    @pytest.mark.parametrize(
        "resource_cls,kwargs",
        [
            (res.Database, {"name": "test"}),
            (res.Schema, {"name": "test", "database": "db"}),
            (res.Role, {"name": "test"}),
            (res.Warehouse, {"name": "test"}),
        ],
    )
    def test_resource_has_resource_type(self, resource_cls, kwargs):
        """Test that all resources have a resource_type property."""
        resource = resource_cls(**kwargs)
        assert resource.resource_type is not None
        assert isinstance(resource.resource_type, ResourceType)

    @pytest.mark.parametrize(
        "resource_cls,kwargs",
        [
            (res.Database, {"name": "test"}),
            (res.Schema, {"name": "test", "database": "db"}),
            (res.Role, {"name": "test"}),
            (res.Warehouse, {"name": "test"}),
        ],
    )
    def test_resource_to_dict(self, resource_cls, kwargs):
        """Test that all resources can be serialized to dict."""
        resource = resource_cls(**kwargs)
        data = resource.to_dict()
        assert isinstance(data, dict)
        assert "name" in data


class TestResourceNameCasing:
    """Tests for resource name case handling."""

    def test_unquoted_name_uppercase(self):
        """Test that unquoted names are uppercased."""
        db = res.Database(name="test_db")
        assert str(db.name) == "TEST_DB"

    def test_quoted_name_preserved(self):
        """Test that quoted names preserve case."""
        db = res.Database(name='"TestDb"')
        assert str(db.name) == '"TestDb"'

    def test_special_chars_quoted(self):
        """Test that special characters cause quoting."""
        db = res.Database(name="test-db")
        # Names with special chars should be quoted
        assert '"' in str(db.to_dict()["name"]) or str(db.name).startswith('"')


class TestResourceValidation:
    """Tests for resource validation rules."""

    def test_warehouse_size_validation(self):
        """Test Warehouse size validation."""
        # Valid size
        wh = res.Warehouse(name="test", warehouse_size="XSMALL")
        assert wh._data.warehouse_size == WarehouseSize.XSMALL

    def test_resource_name_type_validation(self):
        """Test that non-string names are rejected."""
        with pytest.raises(TypeError):
            res.Database(name=123)

    def test_grant_requires_target(self):
        """Test that Grant requires an on_* target."""
        # This should work - has on_database
        grant = res.Grant(priv="USAGE", on_database="test_db", to="role")
        assert grant is not None


class TestResourceDefaults:
    """Tests for resource default values."""

    def test_database_defaults(self):
        """Test Database property defaults."""
        db = res.Database(name="test")
        assert db._data.transient is False
        # data_retention and max_data_extension default to None (use Snowflake defaults)
        assert db._data.data_retention_time_in_days is None
        assert db._data.max_data_extension_time_in_days is None

    def test_schema_defaults(self):
        """Test Schema property defaults."""
        schema = res.Schema(name="test", database="db")
        assert schema._data.transient is False
        # data_retention defaults to None (inherit from database)
        assert schema._data.data_retention_time_in_days is None

    def test_warehouse_defaults(self):
        """Test Warehouse property defaults."""
        wh = res.Warehouse(name="test")
        assert wh._data.warehouse_size == WarehouseSize.XSMALL
        assert wh._data.auto_resume is True
        assert wh._data.auto_suspend == 600

    def test_role_defaults(self):
        """Test Role property defaults."""
        role = res.Role(name="test")
        assert role._data.owner.name == "USERADMIN"


class TestResourceTags:
    """Tests for resource tagging."""

    def test_database_with_tags(self):
        """Test Database with tags."""
        db = res.Database(name="test", tags={"env": "prod", "team": "data"})
        assert db.tags is not None
        assert db.tags.to_dict() == {"env": "prod", "team": "data"}

    def test_schema_with_tags(self):
        """Test Schema with tags."""
        schema = res.Schema(name="test", database="db", tags={"env": "dev"})
        assert schema.tags is not None

    def test_warehouse_with_tags(self):
        """Test Warehouse with tags."""
        wh = res.Warehouse(name="test", tags={"cost_center": "123"})
        assert wh.tags is not None


class TestResourceContainment:
    """Tests for resource containment relationships."""

    def test_schema_contained_in_database(self):
        """Test Schema is contained in Database."""
        db = res.Database(name="my_db")
        schema = res.Schema(name="my_schema", database=db)
        assert schema.container is not None
        assert schema.container.name == "my_db"

    def test_table_contained_in_schema(self):
        """Test Table is contained in Schema."""
        table = res.Table(
            name="my_table",
            database="my_db",
            schema="my_schema",
            columns=[{"name": "id", "data_type": "INT"}],
        )
        assert table.container is not None
        assert table.container.name == "my_schema"
        assert table.container.container.name == "my_db"

    def test_view_contained_in_schema(self):
        """Test View is contained in Schema."""
        view = res.View(
            name="my_view",
            database="my_db",
            schema="my_schema",
            as_="SELECT 1",
        )
        assert view.container is not None
        assert view.container.name == "my_schema"



"""
Unit tests for snowcap/data_provider.py

Tests the data provider functions without requiring a real Snowflake connection.
All functions are tested with mocked execute() results.
"""

import pytest
from unittest.mock import MagicMock, patch, PropertyMock

from snowcap.data_provider import (
    # Helper functions
    _quote_snowflake_identifier,
    _get_owner_identifier,
    _desc_result_to_dict,
    _desc_type2_result_to_dict,
    _desc_type3_result_to_dict,
    _desc_type4_result_to_dict,
    _fail_if_not_granted,
    _filter_result,
    _convert_to_gmt,
    _parse_cluster_keys,
    _parse_function_arguments,
    _parse_function_arguments_2023_compat,
    _parse_list_property,
    _parse_signature,
    _parse_comma_separated_values,
    _parse_packages,
    _parse_storage_location,
    _cast_param_value,
    params_result_to_dict,
    options_result_to_list,
    remove_none_values,
    # Dispatcher functions
    fetch_resource,
    list_resource,
    list_account_scoped_resource,
    list_schema_scoped_resource,
    # Session functions
    fetch_account_locator,
    fetch_region,
)
from snowcap.identifiers import FQN, URN
from snowcap.resource_name import ResourceName
from snowcap.enums import ResourceType

import datetime
import pytz


class TestQuoteSnowflakeIdentifier:
    """Tests for _quote_snowflake_identifier helper function."""

    def test_uppercase_identifier(self):
        # Uppercase identifiers pass through without quotes
        result = _quote_snowflake_identifier("MY_TABLE")
        assert str(result) == "MY_TABLE"

    def test_lowercase_identifier_gets_quoted(self):
        # Lowercase identifiers from Snowflake metadata get quoted
        result = _quote_snowflake_identifier("my_table")
        assert str(result) == '"my_table"'

    def test_identifier_with_special_chars(self):
        result = _quote_snowflake_identifier("my-table")
        assert str(result) == '"my-table"'

    def test_already_quoted_identifier(self):
        result = _quote_snowflake_identifier('"MyTable"')
        assert str(result) == '"MyTable"'

    def test_string_input(self):
        # resource_name_from_snowflake_metadata requires string input
        result = _quote_snowflake_identifier("MY_SCHEMA")
        assert str(result) == "MY_SCHEMA"


class TestGetOwnerIdentifier:
    """Tests for _get_owner_identifier helper function."""

    def test_simple_owner(self):
        data = {"owner": "ACCOUNTADMIN"}
        result = _get_owner_identifier(data)
        assert result == "ACCOUNTADMIN"

    def test_database_role_owner(self):
        # Lowercase names from Snowflake metadata get quoted
        data = {
            "owner": "my_role",
            "owner_role_type": "DATABASE_ROLE",
            "database_name": "my_db"
        }
        result = _get_owner_identifier(data)
        assert result == '"my_db"."my_role"'

    def test_database_role_owner_uppercase(self):
        # Uppercase names pass through without quotes
        data = {
            "owner": "MY_ROLE",
            "owner_role_type": "DATABASE_ROLE",
            "database_name": "MY_DB"
        }
        result = _get_owner_identifier(data)
        assert result == "MY_DB.MY_ROLE"

    def test_role_type_owner(self):
        data = {
            "owner": "SYSADMIN",
            "owner_role_type": "ROLE"
        }
        result = _get_owner_identifier(data)
        assert result == "SYSADMIN"

    def test_empty_owner_with_role_type(self):
        data = {
            "owner": "",
            "owner_role_type": "ROLE"
        }
        result = _get_owner_identifier(data)
        assert result == ""

    def test_unsupported_owner_role_type_raises(self):
        data = {
            "owner": "my_role",
            "owner_role_type": "UNKNOWN_TYPE",
            "database_name": "my_db"
        }
        with pytest.raises(Exception, match="Unsupported owner role type"):
            _get_owner_identifier(data)


class TestDescResultToDict:
    """Tests for _desc_result_to_dict helper function."""

    def test_basic_desc_result(self):
        desc_result = [
            {"property": "NAME", "value": "test_table"},
            {"property": "TYPE", "value": "TABLE"},
        ]
        result = _desc_result_to_dict(desc_result)
        assert result == {"NAME": "test_table", "TYPE": "TABLE"}

    def test_lower_properties(self):
        desc_result = [
            {"property": "NAME", "value": "test_table"},
            {"property": "TYPE", "value": "TABLE"},
        ]
        result = _desc_result_to_dict(desc_result, lower_properties=True)
        assert result == {"name": "test_table", "type": "TABLE"}

    def test_empty_result(self):
        result = _desc_result_to_dict([])
        assert result == {}


class TestDescType2ResultToDict:
    """Tests for _desc_type2_result_to_dict helper function."""

    def test_boolean_property(self):
        desc_result = [
            {"property": "ENABLED", "property_value": "true", "property_type": "Boolean"}
        ]
        result = _desc_type2_result_to_dict(desc_result)
        assert result["ENABLED"] is True

    def test_boolean_false(self):
        desc_result = [
            {"property": "ENABLED", "property_value": "false", "property_type": "Boolean"}
        ]
        result = _desc_type2_result_to_dict(desc_result)
        assert result["ENABLED"] is False

    def test_long_property(self):
        desc_result = [
            {"property": "SIZE", "property_value": "1024", "property_type": "Long"}
        ]
        result = _desc_type2_result_to_dict(desc_result)
        assert result["SIZE"] == "1024"

    def test_long_empty_value(self):
        desc_result = [
            {"property": "SIZE", "property_value": "", "property_type": "Long"}
        ]
        result = _desc_type2_result_to_dict(desc_result)
        assert result["SIZE"] is None

    def test_integer_property(self):
        desc_result = [
            {"property": "COUNT", "property_value": "42", "property_type": "Integer"}
        ]
        result = _desc_type2_result_to_dict(desc_result)
        assert result["COUNT"] == 42

    def test_string_property(self):
        desc_result = [
            {"property": "NAME", "property_value": "my_name", "property_type": "String"}
        ]
        result = _desc_type2_result_to_dict(desc_result)
        assert result["NAME"] == "my_name"

    def test_string_empty_value(self):
        desc_result = [
            {"property": "NAME", "property_value": "", "property_type": "String"}
        ]
        result = _desc_type2_result_to_dict(desc_result)
        assert result["NAME"] is None

    def test_list_property(self):
        desc_result = [
            {"property": "ROLES", "property_value": "[role1, role2]", "property_type": "List"}
        ]
        result = _desc_type2_result_to_dict(desc_result)
        assert result["ROLES"] == ["role1", "role2"]

    def test_object_property(self):
        desc_result = [
            {"property": "CONFIG", "property_value": "[a, b, c]", "property_type": "Object"}
        ]
        result = _desc_type2_result_to_dict(desc_result)
        assert result["CONFIG"] == ["a", "b", "c"]


class TestDescType3ResultToDict:
    """Tests for _desc_type3_result_to_dict helper function."""

    def test_flat_property(self):
        desc_result = [
            {"parent_property": "", "property": "NAME", "property_value": "test", "property_type": "String"}
        ]
        result = _desc_type3_result_to_dict(desc_result)
        assert result["NAME"] == "test"

    def test_nested_property(self):
        desc_result = [
            {"parent_property": "CONFIG", "property": "SIZE", "property_value": "10", "property_type": "Integer"}
        ]
        result = _desc_type3_result_to_dict(desc_result)
        assert result["CONFIG"]["SIZE"] == 10

    def test_multiple_nested_properties(self):
        desc_result = [
            {"parent_property": "CONFIG", "property": "SIZE", "property_value": "10", "property_type": "Integer"},
            {"parent_property": "CONFIG", "property": "ENABLED", "property_value": "true", "property_type": "Boolean"},
        ]
        result = _desc_type3_result_to_dict(desc_result)
        assert result["CONFIG"]["SIZE"] == 10
        assert result["CONFIG"]["ENABLED"] is True


class TestDescType4ResultToDict:
    """Tests for _desc_type4_result_to_dict helper function."""

    def test_basic_result(self):
        desc_result = [
            {"name": "param1", "value": "value1"},
            {"name": "param2", "value": "value2"},
        ]
        result = _desc_type4_result_to_dict(desc_result)
        assert result == {"param1": "value1", "param2": "value2"}

    def test_lower_properties(self):
        desc_result = [
            {"name": "PARAM1", "value": "value1"},
        ]
        result = _desc_type4_result_to_dict(desc_result, lower_properties=True)
        assert result == {"param1": "value1"}


class TestFailIfNotGranted:
    """Tests for _fail_if_not_granted helper function."""

    def test_empty_result_raises(self):
        with pytest.raises(Exception, match="Failed to create grant"):
            _fail_if_not_granted([])

    def test_insufficient_privileges_raises(self):
        result = [{"status": "Grant not executed: Insufficient privileges."}]
        with pytest.raises(Exception, match="Insufficient privileges"):
            _fail_if_not_granted(result)

    def test_success_does_not_raise(self):
        result = [{"status": "Grant succeeded."}]
        _fail_if_not_granted(result)  # Should not raise


class TestFilterResult:
    """Tests for _filter_result helper function."""

    def test_filter_by_single_key(self):
        result = [
            {"name": "TABLE1", "type": "TABLE"},
            {"name": "TABLE2", "type": "VIEW"},
        ]
        filtered = _filter_result(result, type="TABLE")
        assert len(filtered) == 1
        assert filtered[0]["name"] == "TABLE1"

    def test_filter_by_name_resource_name(self):
        result = [
            {"name": "MY_TABLE", "type": "TABLE"},
            {"name": "OTHER_TABLE", "type": "TABLE"},
        ]
        filtered = _filter_result(result, name="my_table")
        assert len(filtered) == 1
        assert filtered[0]["name"] == "MY_TABLE"

    def test_filter_by_multiple_keys(self):
        result = [
            {"name": "TABLE1", "database_name": "DB1", "type": "TABLE"},
            {"name": "TABLE1", "database_name": "DB2", "type": "TABLE"},
            {"name": "TABLE2", "database_name": "DB1", "type": "VIEW"},
        ]
        filtered = _filter_result(result, name="TABLE1", database_name="DB1")
        assert len(filtered) == 1
        assert filtered[0]["database_name"] == "DB1"

    def test_filter_ignores_none_values(self):
        result = [
            {"name": "TABLE1", "type": "TABLE"},
            {"name": "TABLE2", "type": "VIEW"},
        ]
        filtered = _filter_result(result, name=None, type="TABLE")
        assert len(filtered) == 1


class TestConvertToGmt:
    """Tests for _convert_to_gmt helper function."""

    def test_convert_pst_to_gmt(self):
        pst = pytz.timezone("America/Los_Angeles")
        dt = pst.localize(datetime.datetime(2024, 1, 15, 12, 0, 0))
        result = _convert_to_gmt(dt)
        assert result == "2024-01-15 20:00:00"

    def test_none_input_returns_none(self):
        result = _convert_to_gmt(None)
        assert result is None

    def test_custom_format(self):
        pst = pytz.timezone("America/Los_Angeles")
        dt = pst.localize(datetime.datetime(2024, 1, 15, 12, 30, 45))
        result = _convert_to_gmt(dt, fmt_str="%Y-%m-%d %H:%M")
        assert result == "2024-01-15 20:30"


class TestParseClusterKeys:
    """Tests for _parse_cluster_keys helper function."""

    def test_simple_cluster_keys(self):
        result = _parse_cluster_keys("LINEAR(C1, C3)")
        assert result == ["C1", "C3"]

    def test_expression_cluster_keys(self):
        # Note: The function uses simple comma split, so nested expressions get split
        result = _parse_cluster_keys("LINEAR(SUBSTRING(C2, 5, 15), CAST(C1 AS DATE))")
        # This shows the current behavior - nested commas are also split
        assert len(result) > 0
        assert "SUBSTRING(C2" in result[0]

    def test_none_returns_none(self):
        result = _parse_cluster_keys(None)
        assert result is None

    def test_empty_string_returns_none(self):
        result = _parse_cluster_keys("")
        assert result is None


class TestParseFunctionArguments:
    """Tests for _parse_function_arguments helper function."""

    def test_simple_function(self):
        identifier, returns = _parse_function_arguments("FETCH_DATABASE(VARCHAR) RETURN OBJECT")
        # The FQN name field holds just the function name, sig is in the FQN
        assert str(identifier.name) == "FETCH_DATABASE"
        assert returns == "OBJECT"

    def test_multiple_arguments(self):
        identifier, returns = _parse_function_arguments("MY_FUNC(VARCHAR, NUMBER) RETURN TABLE")
        assert str(identifier.name) == "MY_FUNC"
        assert returns == "TABLE"

    def test_returns_fqn_with_arg_types(self):
        identifier, returns = _parse_function_arguments("FETCH_DATABASE(VARCHAR) RETURN OBJECT")
        # The FQN should contain the argument types
        assert identifier.arg_types == ["VARCHAR"]


class TestParseFunctionArguments2023Compat:
    """Tests for _parse_function_arguments_2023_compat helper function."""

    def test_optional_arguments(self):
        identifier, returns = _parse_function_arguments_2023_compat("FETCH_DATABASE(OBJECT [, BOOLEAN]) RETURN OBJECT")
        # Optional brackets are removed
        assert str(identifier.name) == "FETCH_DATABASE"
        assert identifier.arg_types == ["OBJECT", "BOOLEAN"]
        assert returns == "OBJECT"


class TestParseListProperty:
    """Tests for _parse_list_property helper function."""

    def test_simple_list(self):
        result = _parse_list_property("[a, b, c]")
        assert result == ["a", "b", "c"]

    def test_empty_list(self):
        result = _parse_list_property("[]")
        assert result == []

    def test_none_returns_none(self):
        result = _parse_list_property(None)
        assert result is None

    def test_empty_string_returns_none(self):
        result = _parse_list_property("")
        assert result is None

    def test_trims_whitespace(self):
        result = _parse_list_property("[ item1 ,  item2 ]")
        assert result == ["item1", "item2"]


class TestParseSignature:
    """Tests for _parse_signature helper function."""

    def test_simple_signature(self):
        result = _parse_signature("(col1 VARCHAR, col2 NUMBER)")
        assert len(result) == 2

    def test_empty_signature(self):
        result = _parse_signature("()")
        assert result == []


class TestParseCommaSeparatedValues:
    """Tests for _parse_comma_separated_values helper function."""

    def test_simple_values(self):
        result = _parse_comma_separated_values("a, b, c")
        assert result == ["a", "b", "c"]

    def test_none_returns_none(self):
        result = _parse_comma_separated_values(None)
        assert result is None

    def test_empty_string_returns_none(self):
        result = _parse_comma_separated_values("")
        assert result is None


class TestParsePackages:
    """Tests for _parse_packages helper function."""

    def test_simple_packages(self):
        result = _parse_packages("['numpy', 'pandas']")
        assert result == ["numpy", "pandas"]

    def test_none_returns_none(self):
        result = _parse_packages(None)
        assert result is None

    def test_empty_string_returns_none(self):
        result = _parse_packages("")
        assert result is None


class TestParseStorageLocation:
    """Tests for _parse_storage_location helper function."""

    def test_s3_storage_location(self):
        storage_str = '{"name": "loc1", "storage_provider": "S3", "storage_base_url": "s3://bucket/path", "storage_aws_role_arn": "arn:aws:iam::123:role/test"}'
        result = _parse_storage_location(storage_str)
        assert result["name"] == "loc1"
        assert result["storage_provider"] == "S3"
        assert result["storage_base_url"] == "s3://bucket/path"

    def test_with_encryption(self):
        storage_str = '{"name": "loc1", "storage_provider": "S3", "encryption_type": "SSE_S3"}'
        result = _parse_storage_location(storage_str)
        assert result["encryption"]["type"] == "SSE_S3"

    def test_none_returns_none(self):
        result = _parse_storage_location(None)
        assert result is None

    def test_empty_string_returns_none(self):
        result = _parse_storage_location("")
        assert result is None


class TestCastParamValue:
    """Tests for _cast_param_value helper function."""

    def test_boolean_true(self):
        result = _cast_param_value("true", "BOOLEAN")
        assert result is True

    def test_boolean_false(self):
        result = _cast_param_value("false", "BOOLEAN")
        assert result is False

    def test_number_integer(self):
        result = _cast_param_value("42", "NUMBER")
        assert result == 42
        assert isinstance(result, int)

    def test_number_float(self):
        result = _cast_param_value("3.14", "NUMBER")
        assert result == 3.14
        assert isinstance(result, float)

    def test_string(self):
        result = _cast_param_value("hello", "STRING")
        assert result == "hello"

    def test_string_empty(self):
        result = _cast_param_value("", "STRING")
        assert result is None

    def test_unknown_type_returns_raw(self):
        result = _cast_param_value("value", "UNKNOWN")
        assert result == "value"

    def test_invalid_number_raises(self):
        with pytest.raises(Exception, match="Unsupported number type"):
            _cast_param_value("not_a_number", "NUMBER")


class TestParamsResultToDict:
    """Tests for params_result_to_dict helper function."""

    def test_basic_params(self):
        params_result = [
            {"key": "PARAM1", "value": "true", "type": "BOOLEAN"},
            {"key": "PARAM2", "value": "42", "type": "NUMBER"},
            {"key": "PARAM3", "value": "hello", "type": "STRING"},
        ]
        result = params_result_to_dict(params_result)
        assert result["param1"] is True
        assert result["param2"] == 42
        assert result["param3"] == "hello"


class TestOptionsResultToList:
    """Tests for options_result_to_list helper function."""

    def test_simple_options(self):
        result = options_result_to_list("option1, option2, option3")
        assert result == ["option1", "option2", "option3"]


class TestRemoveNoneValues:
    """Tests for remove_none_values helper function."""

    def test_removes_none_at_top_level(self):
        d = {"a": 1, "b": None, "c": "hello"}
        result = remove_none_values(d)
        assert result == {"a": 1, "c": "hello"}

    def test_removes_none_in_nested_dict(self):
        d = {"a": {"b": 1, "c": None}}
        result = remove_none_values(d)
        assert result == {"a": {"b": 1}}

    def test_removes_none_in_list_of_dicts(self):
        d = {"items": [{"a": 1, "b": None}, {"c": 2}]}
        result = remove_none_values(d)
        assert result["items"][0] == {"a": 1}
        assert result["items"][1] == {"c": 2}

    def test_keeps_non_none_values(self):
        d = {"a": 0, "b": False, "c": ""}
        result = remove_none_values(d)
        assert result == {"a": 0, "b": False, "c": ""}


class TestFetchResource:
    """Tests for fetch_resource dispatcher function."""

    @patch("snowcap.data_provider.fetch_database")
    def test_dispatches_to_correct_fetch_function(self, mock_fetch_database):
        mock_fetch_database.return_value = {"name": "MY_DB"}
        mock_session = MagicMock()

        urn = URN(
            resource_type=ResourceType.DATABASE,
            account_locator="ABC123",
            fqn=FQN(name=ResourceName("MY_DB"))
        )

        result = fetch_resource(mock_session, urn)

        mock_fetch_database.assert_called_once_with(mock_session, urn.fqn)
        assert result == {"name": "MY_DB"}

    @patch("snowcap.data_provider.fetch_schema")
    def test_dispatches_to_schema_fetch(self, mock_fetch_schema):
        mock_fetch_schema.return_value = {"name": "MY_SCHEMA"}
        mock_session = MagicMock()

        urn = URN(
            resource_type=ResourceType.SCHEMA,
            account_locator="ABC123",
            fqn=FQN(database=ResourceName("MY_DB"), name=ResourceName("MY_SCHEMA"))
        )

        result = fetch_resource(mock_session, urn)

        mock_fetch_schema.assert_called_once()
        assert result == {"name": "MY_SCHEMA"}

    @patch("snowcap.data_provider.fetch_role")
    def test_returns_none_on_does_not_exist_error(self, mock_fetch_role):
        from snowflake.connector.errors import ProgrammingError
        mock_fetch_role.side_effect = ProgrammingError(errno=2003)
        mock_session = MagicMock()

        urn = URN(
            resource_type=ResourceType.ROLE,
            account_locator="ABC123",
            fqn=FQN(name=ResourceName("MISSING_ROLE"))
        )

        result = fetch_resource(mock_session, urn)
        assert result is None

    @patch("snowcap.data_provider.fetch_role")
    def test_raises_other_programming_errors(self, mock_fetch_role):
        from snowflake.connector.errors import ProgrammingError
        mock_fetch_role.side_effect = ProgrammingError(errno=1234)
        mock_session = MagicMock()

        urn = URN(
            resource_type=ResourceType.ROLE,
            account_locator="ABC123",
            fqn=FQN(name=ResourceName("MY_ROLE"))
        )

        with pytest.raises(ProgrammingError):
            fetch_resource(mock_session, urn)


class TestListResource:
    """Tests for list_resource dispatcher function."""

    @patch("snowcap.data_provider.list_databases")
    def test_dispatches_to_correct_list_function(self, mock_list_databases):
        mock_list_databases.return_value = [FQN(name=ResourceName("DB1")), FQN(name=ResourceName("DB2"))]
        mock_session = MagicMock()

        result = list_resource(mock_session, "database")

        mock_list_databases.assert_called_once_with(mock_session)
        assert len(result) == 2

    @patch("snowcap.data_provider.list_schemas")
    def test_dispatches_to_schemas_list(self, mock_list_schemas):
        mock_list_schemas.return_value = [FQN(database=ResourceName("DB1"), name=ResourceName("S1"))]
        mock_session = MagicMock()

        result = list_resource(mock_session, "schema")

        mock_list_schemas.assert_called_once_with(mock_session)
        assert len(result) == 1

    @patch("snowcap.data_provider.list_tables")
    def test_pluralizes_resource_label(self, mock_list_tables):
        mock_list_tables.return_value = []
        mock_session = MagicMock()

        list_resource(mock_session, "table")

        mock_list_tables.assert_called_once()


class TestListAccountScopedResource:
    """Tests for list_account_scoped_resource helper function."""

    @patch("snowcap.data_provider.execute")
    def test_lists_account_scoped_resources(self, mock_execute):
        mock_execute.return_value = [
            {"name": "RESOURCE1"},
            {"name": "RESOURCE2"},
        ]
        mock_session = MagicMock()

        result = list_account_scoped_resource(mock_session, "WAREHOUSES")

        mock_execute.assert_called_once_with(mock_session, "SHOW WAREHOUSES", cacheable=True)
        assert len(result) == 2
        assert str(result[0].name) == "RESOURCE1"
        assert str(result[1].name) == "RESOURCE2"


class TestListSchemaScopedResource:
    """Tests for list_schema_scoped_resource helper function."""

    @patch("snowcap.data_provider.execute")
    def test_lists_schema_scoped_resources(self, mock_execute):
        mock_execute.return_value = [
            {"name": "TABLE1", "database_name": "DB1", "schema_name": "SCHEMA1"},
            {"name": "TABLE2", "database_name": "DB1", "schema_name": "SCHEMA1"},
        ]
        mock_session = MagicMock()

        result = list_schema_scoped_resource(mock_session, "TABLES")

        mock_execute.assert_called_once_with(mock_session, "SHOW TABLES IN ACCOUNT", cacheable=True)
        assert len(result) == 2

    @patch("snowcap.data_provider.execute")
    def test_filters_system_databases(self, mock_execute):
        mock_execute.return_value = [
            {"name": "TABLE1", "database_name": "MY_DB", "schema_name": "SCHEMA1"},
            {"name": "SYS_TABLE", "database_name": "SNOWFLAKE", "schema_name": "ACCOUNT_USAGE"},
        ]
        mock_session = MagicMock()

        result = list_schema_scoped_resource(mock_session, "TABLES")

        assert len(result) == 1
        assert str(result[0].database) == "MY_DB"


class TestFetchAccountLocator:
    """Tests for fetch_account_locator function."""

    @patch("snowcap.data_provider.execute")
    def test_returns_account_locator(self, mock_execute):
        mock_execute.return_value = [{"ACCOUNT_LOCATOR": "ABC123"}]
        mock_session = MagicMock()

        result = fetch_account_locator(mock_session)

        assert result == "ABC123"


class TestFetchRegion:
    """Tests for fetch_region function."""

    @patch("snowcap.data_provider.execute")
    def test_returns_region(self, mock_execute):
        mock_execute.return_value = [{"CURRENT_REGION()": "AWS_US_WEST_2"}]
        mock_session = MagicMock()

        result = fetch_region(mock_session)

        assert result == {"CURRENT_REGION()": "AWS_US_WEST_2"}


# ================================
# ACCOUNT_USAGE Unit Tests
# ================================


class TestHasAccountUsageAccess:
    """Tests for _has_account_usage_access function."""

    def setup_method(self):
        """Clear caches before each test."""
        from snowcap.data_provider import (
            _ACCOUNT_USAGE_ACCESS_CACHE,
            _ACCOUNT_USAGE_FALLBACK_CACHE,
        )
        _ACCOUNT_USAGE_ACCESS_CACHE.clear()
        _ACCOUNT_USAGE_FALLBACK_CACHE.clear()

    @patch("snowcap.data_provider.execute")
    def test_returns_true_when_access_granted(self, mock_execute):
        """When ACCOUNT_USAGE query succeeds, function returns True."""
        from snowcap.data_provider import _has_account_usage_access
        mock_execute.return_value = [{"1": 1}]  # Query succeeds
        mock_session = MagicMock()

        result = _has_account_usage_access(mock_session)

        assert result is True
        mock_execute.assert_called_once()
        assert "SNOWFLAKE.ACCOUNT_USAGE.GRANTS_TO_ROLES" in mock_execute.call_args[0][1]

    @patch("snowcap.data_provider.execute")
    def test_returns_false_on_access_control_error(self, mock_execute):
        """When ACCOUNT_USAGE query fails with ACCESS_CONTROL_ERR, returns False."""
        from snowflake.connector.errors import ProgrammingError
        from snowcap.data_provider import _has_account_usage_access
        from snowcap.client import ACCESS_CONTROL_ERR

        mock_execute.side_effect = ProgrammingError(errno=ACCESS_CONTROL_ERR)
        mock_session = MagicMock()

        result = _has_account_usage_access(mock_session)

        assert result is False

    @patch("snowcap.data_provider.execute")
    def test_raises_on_other_programming_errors(self, mock_execute):
        """Other ProgrammingErrors should be re-raised."""
        from snowflake.connector.errors import ProgrammingError
        from snowcap.data_provider import _has_account_usage_access

        mock_execute.side_effect = ProgrammingError(errno=1234)
        mock_session = MagicMock()

        with pytest.raises(ProgrammingError):
            _has_account_usage_access(mock_session)

    @patch("snowcap.data_provider.execute")
    def test_caches_result_per_session(self, mock_execute):
        """Result should be cached per session to avoid repeated queries."""
        from snowcap.data_provider import _has_account_usage_access
        mock_execute.return_value = [{"1": 1}]
        mock_session = MagicMock()

        # First call
        result1 = _has_account_usage_access(mock_session)
        # Second call - should use cache
        result2 = _has_account_usage_access(mock_session)

        assert result1 is True
        assert result2 is True
        # Should only be called once due to caching
        assert mock_execute.call_count == 1

    @patch("snowcap.data_provider.execute")
    def test_different_sessions_have_independent_cache(self, mock_execute):
        """Different sessions should have independent cache entries."""
        from snowcap.data_provider import _has_account_usage_access
        mock_execute.return_value = [{"1": 1}]
        session1 = MagicMock()
        session2 = MagicMock()

        _has_account_usage_access(session1)
        _has_account_usage_access(session2)

        # Both sessions should trigger their own query
        assert mock_execute.call_count == 2


class TestFetchGrantsFromAccountUsage:
    """Tests for _fetch_grants_from_account_usage function."""

    def setup_method(self):
        """Clear caches before each test."""
        from snowcap.data_provider import (
            _ACCOUNT_USAGE_ACCESS_CACHE,
            _ACCOUNT_USAGE_FALLBACK_CACHE,
        )
        _ACCOUNT_USAGE_ACCESS_CACHE.clear()
        _ACCOUNT_USAGE_FALLBACK_CACHE.clear()

    @patch("snowcap.data_provider.execute")
    def test_returns_normalized_grants(self, mock_execute):
        """Grants should be normalized to match SHOW GRANTS structure."""
        from snowcap.data_provider import _fetch_grants_from_account_usage
        from datetime import datetime

        mock_execute.return_value = [
            {
                "CREATED_ON": datetime(2024, 1, 1, 12, 0, 0),
                "PRIVILEGE": "SELECT",
                "GRANTED_ON": "TABLE",
                "NAME": "MY_DB.MY_SCHEMA.MY_TABLE",
                "GRANTED_TO": "ACCOUNT ROLE",
                "GRANTEE_NAME": "MY_ROLE",
                "GRANT_OPTION": True,
                "GRANTED_BY": "SYSADMIN",
            }
        ]
        mock_session = MagicMock()

        result = _fetch_grants_from_account_usage(mock_session)

        assert result is not None
        assert len(result) == 1
        grant = result[0]
        # Check lowercase keys
        assert "privilege" in grant
        assert "granted_on" in grant
        # Check ACCOUNT ROLE -> ROLE conversion
        assert grant["granted_to"] == "ROLE"
        # Check boolean -> string conversion
        assert grant["grant_option"] == "true"

    @patch("snowcap.data_provider.execute")
    def test_converts_database_role_granted_to(self, mock_execute):
        """DATABASE_ROLE should be converted to DATABASE ROLE."""
        from snowcap.data_provider import _fetch_grants_from_account_usage
        from datetime import datetime

        mock_execute.return_value = [
            {
                "CREATED_ON": datetime(2024, 1, 1, 12, 0, 0),
                "PRIVILEGE": "USAGE",
                "GRANTED_ON": "DATABASE",
                "NAME": "MY_DB",
                "GRANTED_TO": "DATABASE_ROLE",
                "GRANTEE_NAME": "MY_DB.MY_DB_ROLE",
                "GRANT_OPTION": False,
                "GRANTED_BY": "SYSADMIN",
            }
        ]
        mock_session = MagicMock()

        result = _fetch_grants_from_account_usage(mock_session)

        assert result is not None
        assert result[0]["granted_to"] == "DATABASE ROLE"
        assert result[0]["grant_option"] == "false"

    @patch("snowcap.data_provider.execute")
    def test_returns_none_on_access_control_error(self, mock_execute):
        """Returns None on permission error (signaling fallback needed)."""
        from snowflake.connector.errors import ProgrammingError
        from snowcap.data_provider import _fetch_grants_from_account_usage
        from snowcap.client import ACCESS_CONTROL_ERR

        mock_execute.side_effect = ProgrammingError(errno=ACCESS_CONTROL_ERR)
        mock_session = MagicMock()

        result = _fetch_grants_from_account_usage(mock_session)

        assert result is None

    @patch("snowcap.data_provider.execute")
    def test_returns_none_on_unexpected_error(self, mock_execute):
        """Returns None on unexpected error (signaling fallback needed)."""
        from snowcap.data_provider import _fetch_grants_from_account_usage

        mock_execute.side_effect = Exception("Unexpected error")
        mock_session = MagicMock()

        result = _fetch_grants_from_account_usage(mock_session)

        assert result is None

    @patch("snowcap.data_provider.execute")
    def test_marks_fallback_on_error(self, mock_execute):
        """On error, should mark fallback cache for this session."""
        from snowcap.data_provider import (
            _fetch_grants_from_account_usage,
            _ACCOUNT_USAGE_FALLBACK_CACHE,
        )

        mock_execute.side_effect = Exception("Unexpected error")
        mock_session = MagicMock()

        _fetch_grants_from_account_usage(mock_session)

        # Fallback should be marked
        assert _ACCOUNT_USAGE_FALLBACK_CACHE.get(id(mock_session)) is True


class TestFetchRoleGrantsToUsersFromAccountUsage:
    """Tests for _fetch_role_grants_to_users_from_account_usage function."""

    def setup_method(self):
        """Clear caches before each test."""
        from snowcap.data_provider import (
            _ACCOUNT_USAGE_ACCESS_CACHE,
            _ACCOUNT_USAGE_FALLBACK_CACHE,
        )
        _ACCOUNT_USAGE_ACCESS_CACHE.clear()
        _ACCOUNT_USAGE_FALLBACK_CACHE.clear()

    @patch("snowcap.data_provider.execute")
    def test_returns_normalized_user_grants(self, mock_execute):
        """User grants should be normalized to match SHOW GRANTS OF ROLE structure."""
        from snowcap.data_provider import _fetch_role_grants_to_users_from_account_usage
        from datetime import datetime

        mock_execute.return_value = [
            {
                "CREATED_ON": datetime(2024, 1, 1, 12, 0, 0),
                "ROLE": "MY_ROLE",
                "GRANTED_TO": "USER",
                "GRANTEE_NAME": "MY_USER",
                "GRANTED_BY": "SECURITYADMIN",
            }
        ]
        mock_session = MagicMock()

        result = _fetch_role_grants_to_users_from_account_usage(mock_session)

        assert result is not None
        assert len(result) == 1
        grant = result[0]
        assert "role" in grant
        assert grant["role"] == "MY_ROLE"
        assert grant["granted_to"] == "USER"
        assert grant["grantee_name"] == "MY_USER"

    @patch("snowcap.data_provider.execute")
    def test_returns_none_on_error(self, mock_execute):
        """Returns None on error (signaling fallback needed)."""
        from snowcap.data_provider import _fetch_role_grants_to_users_from_account_usage

        mock_execute.side_effect = Exception("Unexpected error")
        mock_session = MagicMock()

        result = _fetch_role_grants_to_users_from_account_usage(mock_session)

        assert result is None


class TestShouldUseAccountUsage:
    """Tests for _should_use_account_usage helper function."""

    def setup_method(self):
        """Clear caches before each test."""
        from snowcap.data_provider import (
            _ACCOUNT_USAGE_ACCESS_CACHE,
            _ACCOUNT_USAGE_FALLBACK_CACHE,
        )
        _ACCOUNT_USAGE_ACCESS_CACHE.clear()
        _ACCOUNT_USAGE_FALLBACK_CACHE.clear()

    @patch("snowcap.data_provider._has_account_usage_access")
    def test_returns_false_when_config_disabled(self, mock_has_access):
        """Returns False when use_account_usage config is False."""
        from snowcap.data_provider import _should_use_account_usage
        mock_session = MagicMock()

        result = _should_use_account_usage(mock_session, use_account_usage=False)

        assert result is False
        # Should not even check access when config is disabled
        mock_has_access.assert_not_called()

    @patch("snowcap.data_provider._has_account_usage_access")
    def test_returns_false_when_fallback_cached(self, mock_has_access):
        """Returns False when session has previous ACCOUNT_USAGE failures."""
        from snowcap.data_provider import (
            _should_use_account_usage,
            _ACCOUNT_USAGE_FALLBACK_CACHE,
        )
        mock_session = MagicMock()
        _ACCOUNT_USAGE_FALLBACK_CACHE[id(mock_session)] = True

        result = _should_use_account_usage(mock_session, use_account_usage=True)

        assert result is False
        # Should not check access when fallback is cached
        mock_has_access.assert_not_called()

    @patch("snowcap.data_provider._has_account_usage_access")
    def test_returns_access_check_result_when_enabled(self, mock_has_access):
        """Returns result of _has_account_usage_access when config is enabled."""
        from snowcap.data_provider import _should_use_account_usage
        mock_session = MagicMock()
        mock_has_access.return_value = True

        result = _should_use_account_usage(mock_session, use_account_usage=True)

        assert result is True
        mock_has_access.assert_called_once_with(mock_session)

    @patch("snowcap.data_provider._has_account_usage_access")
    def test_returns_false_when_no_access(self, mock_has_access):
        """Returns False when session doesn't have ACCOUNT_USAGE access."""
        from snowcap.data_provider import _should_use_account_usage
        mock_session = MagicMock()
        mock_has_access.return_value = False

        result = _should_use_account_usage(mock_session, use_account_usage=True)

        assert result is False


class TestMarkAccountUsageFallback:
    """Tests for _mark_account_usage_fallback function."""

    def setup_method(self):
        """Clear caches before each test."""
        from snowcap.data_provider import _ACCOUNT_USAGE_FALLBACK_CACHE
        _ACCOUNT_USAGE_FALLBACK_CACHE.clear()

    def test_marks_session_for_fallback(self):
        """Should mark session ID in fallback cache."""
        from snowcap.data_provider import (
            _mark_account_usage_fallback,
            _ACCOUNT_USAGE_FALLBACK_CACHE,
        )
        mock_session = MagicMock()

        _mark_account_usage_fallback(mock_session)

        assert _ACCOUNT_USAGE_FALLBACK_CACHE.get(id(mock_session)) is True


class TestFetchRolePrivilegesAccountUsage:
    """Tests for fetch_role_privileges with ACCOUNT_USAGE integration."""

    def setup_method(self):
        """Clear caches before each test."""
        from snowcap.data_provider import (
            _ACCOUNT_USAGE_ACCESS_CACHE,
            _ACCOUNT_USAGE_FALLBACK_CACHE,
        )
        _ACCOUNT_USAGE_ACCESS_CACHE.clear()
        _ACCOUNT_USAGE_FALLBACK_CACHE.clear()

    @patch("snowcap.data_provider._should_use_account_usage")
    @patch("snowcap.data_provider._fetch_grants_from_account_usage")
    @patch("snowcap.data_provider._show_grants_to_role")
    def test_uses_account_usage_when_enabled_and_available(
        self, mock_show_grants, mock_fetch_au, mock_should_use
    ):
        """When ACCOUNT_USAGE is enabled and available, uses ACCOUNT_USAGE."""
        from snowcap.data_provider import fetch_role_privileges
        from datetime import datetime

        mock_should_use.return_value = True
        mock_fetch_au.return_value = [
            {
                "created_on": datetime(2024, 1, 1, 12, 0, 0),
                "privilege": "USAGE",
                "granted_on": "DATABASE",
                "name": "MY_DB",
                "granted_to": "ROLE",
                "grantee_name": "MY_ROLE",
                "grant_option": "false",
                "granted_by": "SYSADMIN",
            }
        ]
        mock_session = MagicMock()
        roles = {"MY_ROLE": MagicMock()}

        result = fetch_role_privileges(mock_session, roles, use_account_usage=True)

        mock_should_use.assert_called_once()
        mock_fetch_au.assert_called_once()
        # SHOW GRANTS should not be called when ACCOUNT_USAGE succeeds
        mock_show_grants.assert_not_called()
        assert "MY_ROLE" in result

    @patch("snowcap.data_provider._should_use_account_usage")
    @patch("snowcap.data_provider._show_grants_to_role")
    def test_falls_back_to_show_when_disabled(self, mock_show_grants, mock_should_use):
        """When ACCOUNT_USAGE is disabled, uses SHOW GRANTS."""
        from snowcap.data_provider import fetch_role_privileges

        mock_should_use.return_value = False
        mock_show_grants.return_value = []
        mock_session = MagicMock()
        roles = {"MY_ROLE": MagicMock()}

        fetch_role_privileges(mock_session, roles, use_account_usage=False)

        mock_show_grants.assert_called()

    @patch("snowcap.data_provider._should_use_account_usage")
    @patch("snowcap.data_provider._fetch_grants_from_account_usage")
    @patch("snowcap.data_provider._show_grants_to_role")
    def test_falls_back_when_account_usage_returns_none(
        self, mock_show_grants, mock_fetch_au, mock_should_use
    ):
        """When ACCOUNT_USAGE query fails (returns None), falls back to SHOW."""
        from snowcap.data_provider import fetch_role_privileges

        mock_should_use.return_value = True
        mock_fetch_au.return_value = None  # Signals failure
        mock_show_grants.return_value = []
        mock_session = MagicMock()
        roles = {"MY_ROLE": MagicMock()}

        fetch_role_privileges(mock_session, roles, use_account_usage=True)

        # Both should be called - ACCOUNT_USAGE first, then fallback
        mock_fetch_au.assert_called_once()
        mock_show_grants.assert_called()


class TestBlueprintConfigUseAccountUsage:
    """Tests for use_account_usage config flag in BlueprintConfig."""

    def test_default_is_false(self):
        """use_account_usage should default to False for performance (small manifests)."""
        from snowcap.blueprint_config import BlueprintConfig

        config = BlueprintConfig()

        assert config.use_account_usage is False

    def test_can_be_set_to_false(self):
        """use_account_usage can be explicitly set to False."""
        from snowcap.blueprint_config import BlueprintConfig

        config = BlueprintConfig(use_account_usage=False)

        assert config.use_account_usage is False



import json
import re
from copy import deepcopy

import pytest

from snowcap import resources as res


def strip_ansi(text):
    """Remove ANSI color codes from text."""
    return re.sub(r"\x1b\[[0-9;]*m", "", text)


def find_change_by_urn(plan, urn):
    """Find a change in the plan by its URN."""
    for change in plan:
        if change.urn == urn:
            return change
    return None


def flatten_sql_commands(sql_commands: list[dict]) -> list[str]:
    """Flatten compile_plan_to_sql output to a list of SQL strings for testing."""
    result = ["USE SECONDARY ROLES ALL"]
    last_role = None
    for cmd in sql_commands:
        if cmd["role"] != last_role:
            result.append(f"USE ROLE {cmd['role']}")
            last_role = cmd["role"]
        result.extend(cmd["commands"])
    return result


from snowcap import var
from snowcap.blueprint import (
    Blueprint,
    CreateResource,
    _merge_pointers,
    compile_plan_to_sql,
    diff,
    dump_plan,
)
from snowcap.blueprint_config import BlueprintConfig
from snowcap.enums import AccountEdition, BlueprintScope, ResourceType
from snowcap.exceptions import (
    DuplicateResourceException,
    MissingVarException,
    NonConformingPlanException,
    WrongEditionException,
)
from snowcap.identifiers import FQN, URN, parse_URN
from snowcap.resource_name import ResourceName
from snowcap.resources.resource import ResourcePointer
from snowcap.var import VarString


@pytest.fixture
def session_ctx() -> dict:
    return {
        "account": "SOMEACCT",
        "account_edition": AccountEdition.ENTERPRISE,
        "account_locator": "ABCD123",
        "role": "SYSADMIN",
        "available_roles": [
            "SYSADMIN",
            "USERADMIN",
            "ACCOUNTADMIN",
            "SECURITYADMIN",
            "PUBLIC",
        ],
    }


@pytest.fixture
def remote_state() -> dict:
    return {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
    }


@pytest.fixture
def resource_manifest():
    session_ctx = {
        "account": "SOMEACCT",
        "account_edition": AccountEdition.ENTERPRISE,
        "account_locator": "ABCD123",
        "current_role": "SYSADMIN",
        "available_roles": ["SYSADMIN", "USERADMIN"],
    }
    db = res.Database(name="DB")
    schema = res.Schema(name="SCHEMA", database=db)
    table = res.Table(name="TABLE", columns=[{"name": "ID", "data_type": "INT"}])
    schema.add(table)
    view = res.View(name="VIEW", schema=schema, as_="SELECT 1")
    udf = res.PythonUDF(
        name="SOMEUDF",
        returns="VARCHAR",
        args=[],
        runtime_version="3.9",
        handler="main",
        comment="This is a UDF comment",
    )
    schema.add(udf)
    blueprint = Blueprint(name="blueprint", resources=[db, table, schema, view, udf])
    manifest = blueprint.generate_manifest(session_ctx)
    return manifest


def test_blueprint_with_database(resource_manifest):

    db_urn = parse_URN("urn::ABCD123:database/DB")
    assert db_urn in resource_manifest
    assert resource_manifest[db_urn].data == {
        "name": "DB",
        "owner": "SYSADMIN",
        "comment": None,
        "catalog": None,
        "external_volume": None,
        "data_retention_time_in_days": None,  # Uses Snowflake default
        "default_ddl_collation": None,  # Uses Snowflake default
        "max_data_extension_time_in_days": None,  # Uses Snowflake default
        "transient": False,
    }


def test_blueprint_with_schema(resource_manifest):
    schema_urn = parse_URN("urn::ABCD123:schema/DB.SCHEMA")
    assert schema_urn in resource_manifest
    assert resource_manifest[schema_urn].data == {
        "comment": None,
        "data_retention_time_in_days": None,  # Inherits from database
        "default_ddl_collation": None,  # Inherits from database
        "managed_access": False,
        "max_data_extension_time_in_days": None,  # Inherits from database
        "name": "SCHEMA",
        "owner": "SYSADMIN",
        "transient": False,
    }


def test_blueprint_with_view(resource_manifest):
    view_urn = parse_URN("urn::ABCD123:view/DB.SCHEMA.VIEW")
    assert view_urn in resource_manifest
    assert resource_manifest[view_urn].data == {
        "as_": "SELECT 1",
        "change_tracking": False,
        "columns": None,
        "comment": None,
        "copy_grants": False,
        "name": "VIEW",
        "owner": "SYSADMIN",
        "recursive": None,
        "secure": False,
        "volatile": None,
    }


def test_blueprint_with_table(resource_manifest):
    table_urn = parse_URN("urn::ABCD123:table/DB.SCHEMA.TABLE")
    assert table_urn in resource_manifest
    assert resource_manifest[table_urn].data == {
        "name": "TABLE",
        "owner": "SYSADMIN",
        "columns": [
            {
                "name": "ID",
                "data_type": "NUMBER(38,0)",
                "collate": None,
                "comment": None,
                "constraint": None,
                "not_null": False,
                "default": None,
                "tags": None,
            }
        ],
        "constraints": None,
        "transient": False,
        "cluster_by": None,
        "enable_schema_evolution": False,
        "data_retention_time_in_days": None,
        "max_data_extension_time_in_days": None,
        "change_tracking": False,
        "default_ddl_collation": None,
        "copy_grants": None,
        "row_access_policy": None,
        "comment": None,
    }


def test_blueprint_with_udf(resource_manifest):
    # parse URN is incorrectly stripping the parens. Not sure what the correct behavior should be
    # udf_urn = parse_URN("urn::ABCD123:function/DB.PUBLIC.SOMEUDF()")
    udf_urn = URN(
        resource_type=ResourceType.FUNCTION,
        fqn=FQN(
            database=ResourceName("DB"),
            schema=ResourceName("SCHEMA"),
            name=ResourceName("SOMEUDF"),
            arg_types=[],
        ),
        account_locator="ABCD123",
    )
    assert udf_urn in resource_manifest
    assert resource_manifest[udf_urn].data == {
        "name": "SOMEUDF",
        "owner": "SYSADMIN",
        "returns": "VARCHAR",
        "handler": "main",
        "runtime_version": "3.9",
        "comment": "This is a UDF comment",
        "args": [],
        "as_": None,
        "copy_grants": False,
        "language": "PYTHON",
        "external_access_integrations": None,
        "imports": None,
        "null_handling": None,
        "packages": None,
        "secrets": None,
        "secure": None,
        "volatility": None,
    }


def test_blueprint_resource_owned_by_plan_role(session_ctx, remote_state):
    role = res.Role("SOME_ROLE")
    wh = res.Warehouse("WH", owner=role)
    grant = res.RoleGrant(role=role, to_role="SYSADMIN")
    blueprint = Blueprint(name="blueprint", resources=[wh, role, grant])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)

    # Check all expected URNs are present (order not guaranteed)
    plan_urns = set(change.urn for change in plan)
    assert parse_URN("urn::ABCD123:role/SOME_ROLE") in plan_urns
    assert parse_URN("urn::ABCD123:role_grant/SOME_ROLE?role=SYSADMIN") in plan_urns
    assert parse_URN("urn::ABCD123:warehouse/WH") in plan_urns
    assert len(plan_urns) == 3

    changes = flatten_sql_commands(compile_plan_to_sql(session_ctx, plan))
    # Check expected commands are present (order may vary)
    assert "USE SECONDARY ROLES ALL" in changes
    assert "CREATE ROLE SOME_ROLE" in changes
    assert "GRANT ROLE SOME_ROLE TO ROLE SYSADMIN" in changes
    assert any(c.startswith("CREATE WAREHOUSE WH") for c in changes)
    assert "GRANT OWNERSHIP ON WAREHOUSE WH TO ROLE SOME_ROLE COPY CURRENT GRANTS" in changes


def test_blueprint_deduplicate_resources(session_ctx, remote_state):
    blueprint = Blueprint(
        name="blueprint",
        resources=[
            res.Database("DB"),
            ResourcePointer(name="DB", resource_type=ResourceType.DATABASE),
        ],
    )
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    assert len(plan) == 1
    assert isinstance(plan[0], CreateResource)
    assert plan[0].urn == parse_URN("urn::ABCD123:database/DB")
    assert plan[0].resource_cls == res.Database

    blueprint = Blueprint(
        name="blueprint",
        resources=[
            res.Database("DB"),
            res.Database("DB", comment="This is a comment"),
        ],
    )
    with pytest.raises(DuplicateResourceException):
        blueprint.generate_manifest(session_ctx)

    blueprint = Blueprint(
        name="blueprint",
        resources=[
            res.Grant(priv="USAGE", on_database="DB", to="SOME_ROLE"),
            res.Grant(priv="USAGE", on_database="DB", to="SOME_ROLE"),
        ],
    )
    with pytest.raises(DuplicateResourceException):
        blueprint.generate_manifest(session_ctx)


def test_blueprint_dont_add_public_schema(session_ctx, remote_state):
    db = res.Database("DB")
    public = ResourcePointer(name="PUBLIC", resource_type=ResourceType.SCHEMA)
    blueprint = Blueprint(
        name="blueprint",
        resources=[db, public],
    )
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    assert len(plan) == 1
    assert isinstance(plan[0], CreateResource)
    assert plan[0].urn == parse_URN("urn::ABCD123:database/DB")
    assert plan[0].resource_cls == res.Database


def test_blueprint_implied_container_tree(session_ctx, remote_state):
    remote_state[parse_URN("urn::ABCD123:database/STATIC_DB")] = {"owner": "SYSADMIN"}
    remote_state[parse_URN("urn::ABCD123:schema/STATIC_DB.PUBLIC")] = {"owner": "SYSADMIN"}
    func = res.JavascriptUDF(
        name="func", args=[], returns="INT", as_="return 1;", database="STATIC_DB", schema="public"
    )
    blueprint = Blueprint(name="blueprint", resources=[func])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    assert len(plan) == 1
    assert isinstance(plan[0], CreateResource)
    assert plan[0].urn.fqn.name == "FUNC"
    assert plan[0].resource_cls == res.JavascriptUDF


def test_blueprint_chained_ownership(session_ctx, remote_state):
    role = res.Role("SOME_ROLE")
    role_grant = res.RoleGrant(role=role, to_role="SYSADMIN")
    db = res.Database("DB", owner=role)
    schema = res.Schema("SCHEMA", database=db, owner=role)
    blueprint = Blueprint(name="blueprint", resources=[db, schema, role_grant, role])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    assert len(plan) == 4
    # Find changes by URN instead of relying on order
    role_change = find_change_by_urn(plan, parse_URN("urn::ABCD123:role/SOME_ROLE"))
    grant_change = find_change_by_urn(plan, parse_URN("urn::ABCD123:role_grant/SOME_ROLE?role=SYSADMIN"))
    db_change = find_change_by_urn(plan, parse_URN("urn::ABCD123:database/DB"))
    schema_change = find_change_by_urn(plan, parse_URN("urn::ABCD123:schema/DB.SCHEMA"))
    assert isinstance(role_change, CreateResource)
    assert role_change.resource_cls == res.Role
    assert isinstance(grant_change, CreateResource)
    assert grant_change.resource_cls == res.RoleGrant
    assert isinstance(db_change, CreateResource)
    assert db_change.resource_cls == res.Database
    assert isinstance(schema_change, CreateResource)
    assert schema_change.resource_cls == res.Schema


def test_blueprint_polymorphic_resource_resolution(session_ctx, remote_state):

    role = res.Role(name="DEMO_ROLE")
    sysad_grant = res.RoleGrant(role=role, to_role="SYSADMIN")
    test_db = res.Database(name="TEST_TITAN", transient=False, data_retention_time_in_days=1, comment="Test Titan")
    schema = res.Schema(name="TEST_SCHEMA", database=test_db, transient=False, comment="Test Titan Schema")
    warehouse = res.Warehouse(name="FAKER_LOADER", auto_suspend=60)

    future_schema_grant = res.Grant(priv="usage", on=["FUTURE", "SCHEMAS", test_db], to=role)
    post_grant = [future_schema_grant]

    grants = [
        res.Grant(priv="usage", to=role, on=warehouse),
        res.Grant(priv="operate", to=role, on=warehouse),
        res.Grant(priv="usage", to=role, on=test_db),
        # future_schema_grant,
        # x
        # Grant(priv="usage", to=role, on=schema)
    ]

    sales_table = res.Table(
        name="faker_data",
        schema=schema,
        columns=[
            res.Column(name="NAME", data_type="VARCHAR(16777216)"),
            res.Column(name="EMAIL", data_type="VARCHAR(16777216)"),
            res.Column(name="ADDRESS", data_type="VARCHAR(16777216)"),
            res.Column(name="ORDERED_AT_UTC", data_type="NUMBER(38,0)"),
            res.Column(name="EXTRACTED_AT_UTC", data_type="NUMBER(38,0)"),
            res.Column(name="SALES_ORDER_ID", data_type="VARCHAR(16777216)"),
        ],
        comment="Test Table",
    )
    blueprint = Blueprint(
        name="blueprint",
        resources=[
            role,
            sysad_grant,
            # user_grant,
            test_db,
            # *pre_grant,
            schema,
            sales_table,
            # pipe,
            warehouse,
            *grants,
        ],
    )
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    assert len(plan) == 9


def test_blueprint_scope_sorting(session_ctx, remote_state):
    db = res.Database(name="DB")
    schema = res.Schema(name="SCHEMA", database=db)
    view = res.View(name="SOME_VIEW", schema=schema, as_="SELECT 1")
    blueprint = Blueprint(name="blueprint", resources=[view, schema, db])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    assert len(plan) == 3
    # Find changes by URN instead of relying on order
    db_change = find_change_by_urn(plan, parse_URN("urn::ABCD123:database/DB"))
    schema_change = find_change_by_urn(plan, parse_URN("urn::ABCD123:schema/DB.SCHEMA"))
    view_change = find_change_by_urn(plan, parse_URN("urn::ABCD123:view/DB.SCHEMA.SOME_VIEW"))
    assert isinstance(db_change, CreateResource)
    assert db_change.resource_cls == res.Database
    assert isinstance(schema_change, CreateResource)
    assert schema_change.resource_cls == res.Schema
    assert isinstance(view_change, CreateResource)
    assert view_change.resource_cls == res.View


def test_blueprint_reference_sorting(session_ctx, remote_state):
    db1 = res.Database(name="DB1")
    db2 = res.Database(name="DB2")
    db2.requires(db1)
    db3 = res.Database(name="DB3")
    db3.requires(db2)
    blueprint = Blueprint(resources=[db3, db1, db2])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    assert len(plan) == 3
    # Find changes by URN instead of relying on order
    db1_change = find_change_by_urn(plan, parse_URN("urn::ABCD123:database/DB1"))
    db2_change = find_change_by_urn(plan, parse_URN("urn::ABCD123:database/DB2"))
    db3_change = find_change_by_urn(plan, parse_URN("urn::ABCD123:database/DB3"))
    assert isinstance(db1_change, CreateResource)
    assert db1_change.resource_cls == res.Database
    assert isinstance(db2_change, CreateResource)
    assert db2_change.resource_cls == res.Database
    assert isinstance(db3_change, CreateResource)
    assert db3_change.resource_cls == res.Database


def test_blueprint_ownership_sorting(session_ctx, remote_state):

    role = res.Role(name="SOME_ROLE")
    role_grant = res.RoleGrant(role=role, to_role="SYSADMIN")
    wh = res.Warehouse(name="WH", owner=role)

    blueprint = Blueprint(resources=[wh, role_grant, role])
    manifest = blueprint.generate_manifest(session_ctx)

    plan = diff(remote_state, manifest)
    assert len(plan) == 3
    # Find changes by URN instead of relying on order
    role_change = find_change_by_urn(plan, parse_URN("urn::ABCD123:role/SOME_ROLE"))
    grant_change = find_change_by_urn(plan, parse_URN("urn::ABCD123:role_grant/SOME_ROLE?role=SYSADMIN"))
    wh_change = find_change_by_urn(plan, parse_URN("urn::ABCD123:warehouse/WH"))
    assert isinstance(role_change, CreateResource)
    assert role_change.resource_cls == res.Role
    assert isinstance(grant_change, CreateResource)
    assert grant_change.resource_cls == res.RoleGrant
    assert isinstance(wh_change, CreateResource)
    assert wh_change.resource_cls == res.Warehouse

    sql = flatten_sql_commands(compile_plan_to_sql(session_ctx, plan))
    # Check expected commands are present (order may vary)
    assert "USE SECONDARY ROLES ALL" in sql
    assert "CREATE ROLE SOME_ROLE" in sql
    assert "GRANT ROLE SOME_ROLE TO ROLE SYSADMIN" in sql
    assert any(s.startswith("CREATE WAREHOUSE WH") for s in sql)
    assert "GRANT OWNERSHIP ON WAREHOUSE WH TO ROLE SOME_ROLE COPY CURRENT GRANTS" in sql


def test_blueprint_dump_plan_create(session_ctx, remote_state):
    blueprint = Blueprint(resources=[res.Role("role1")])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    plan_json_str = dump_plan(plan, format="json")
    assert json.loads(plan_json_str) == [
        {
            "action": "CREATE",
            "urn": "urn::ABCD123:role/ROLE1",
            "resource_cls": "Role",
            "container": None,
            "after": {"name": "ROLE1", "owner": "USERADMIN", "comment": None},
        }
    ]
    plan_str = strip_ansi(dump_plan(plan, format="text"))
    assert (
        plan_str
        == """
» snowcap
» Plan: 1 to create, 0 to update, 0 to transfer, 0 to drop.

━━━ ROLES ━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━
+ CREATE: ROLE1 (owner: USERADMIN)

"""
    )


def test_blueprint_dump_plan_update(session_ctx):
    remote_state = {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
        parse_URN("urn::ABCD123:role/ROLE1"): {
            "name": "ROLE1",
            "owner": "USERADMIN",
            "comment": "old",
        },
    }
    blueprint = Blueprint(resources=[res.Role("role1", comment="new")])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    plan_json_str = dump_plan(plan, format="json")
    assert json.loads(plan_json_str) == [
        {
            "action": "UPDATE",
            "resource_cls": "Role",
            "urn": "urn::ABCD123:role/ROLE1",
            "before": {"name": "ROLE1", "owner": "USERADMIN", "comment": "old"},
            "after": {"name": "ROLE1", "owner": "USERADMIN", "comment": "new"},
            "delta": {"comment": "new"},
        }
    ]
    plan_str = strip_ansi(dump_plan(plan, format="text"))
    assert (
        plan_str
        == """
» snowcap
» Plan: 0 to create, 1 to update, 0 to transfer, 0 to drop.

━━━ ROLES ━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━
~ UPDATE: ROLE1
  ┌──────────┬────────┬───────┐
  │ Property │ Before │ After │
  ├──────────┼────────┼───────┤
  │ comment  │ old    │ new   │
  └──────────┴────────┴───────┘

"""
    )


def test_blueprint_dump_plan_transfer(session_ctx):
    remote_state = {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
        parse_URN("urn::ABCD123:role/ROLE1"): {
            "name": "ROLE1",
            "owner": "ACCOUNTADMIN",
            "comment": None,
        },
    }
    blueprint = Blueprint(resources=[res.Role("role1", owner="USERADMIN")])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    plan_json_str = dump_plan(plan, format="json")
    assert json.loads(plan_json_str) == [
        {
            "action": "TRANSFER",
            "resource_cls": "Role",
            "urn": "urn::ABCD123:role/ROLE1",
            "from_owner": "ACCOUNTADMIN",
            "to_owner": "USERADMIN",
        }
    ]
    plan_str = strip_ansi(dump_plan(plan, format="text"))
    assert (
        plan_str
        == """
» snowcap
» Plan: 0 to create, 0 to update, 1 to transfer, 0 to drop.

━━━ ROLES ━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━
~ TRANSFER: ROLE1
  ┌──────────┬──────────────┬───────────┐
  │ Property │ Before       │ After     │
  ├──────────┼──────────────┼───────────┤
  │ owner    │ ACCOUNTADMIN │ USERADMIN │
  └──────────┴──────────────┴───────────┘

"""
    )


def test_blueprint_dump_plan_drop(session_ctx):
    remote_state = {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
        parse_URN("urn::ABCD123:role/ROLE1"): {
            "name": "ROLE1",
            "owner": "ACCOUNTADMIN",
            "comment": None,
        },
    }
    blueprint = Blueprint(resources=[], sync_resources=[ResourceType.ROLE])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    plan_json_str = dump_plan(plan, format="json")
    plan_dict = json.loads(plan_json_str)
    assert len(plan_dict) == 1
    assert plan_dict[0] == {
        "action": "DROP",
        "urn": "urn::ABCD123:role/ROLE1",
        "before": {"name": "ROLE1", "owner": "ACCOUNTADMIN", "comment": None},
    }

    plan_str = strip_ansi(dump_plan(plan, format="text"))
    assert (
        plan_str
        == """
» snowcap
» Plan: 0 to create, 0 to update, 0 to transfer, 1 to drop.

━━━ ROLES ━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━
- DROP:   ROLE1

"""
    )


def test_blueprint_vars(session_ctx):
    blueprint = Blueprint(
        resources=[res.Role(name="role", comment=var.role_comment)],
        vars={"role_comment": "var role comment"},
    )
    manifest = blueprint.generate_manifest(session_ctx)
    assert manifest.resources[1].data["comment"] == "var role comment"

    role = res.Role(name="role", comment="some comment {{ var.suffix }}")
    assert isinstance(role._data.comment, VarString)
    blueprint = Blueprint(
        resources=[role],
        vars={"suffix": "1234"},
    )
    manifest = blueprint.generate_manifest(session_ctx)
    assert manifest.resources[1].data["comment"] == "some comment 1234"

    role = res.Role(name=var.role_name)
    assert isinstance(role.name, VarString)
    blueprint = Blueprint(
        resources=[role],
        vars={"role_name": "role123"},
    )
    manifest = blueprint.generate_manifest(session_ctx)
    assert manifest.resources[1].data["name"] == "role123"

    role = res.Role(name="role_{{ var.suffix }}")
    assert isinstance(role.name, VarString)
    blueprint = Blueprint(
        resources=[role],
        vars={"suffix": "5678"},
    )
    manifest = blueprint.generate_manifest(session_ctx)
    assert manifest.resources[1].data["name"] == "role_5678"


def test_blueprint_vars_spec(session_ctx):
    blueprint = Blueprint(
        resources=[res.Role(name="role", comment=var.role_comment)],
        vars_spec=[
            {
                "name": "role_comment",
                "type": "string",
                "default": "var role comment",
            }
        ],
    )
    assert blueprint._config.vars == {"role_comment": "var role comment"}
    manifest = blueprint.generate_manifest(session_ctx)
    assert manifest.resources[1].data["comment"] == "var role comment"

    with pytest.raises(MissingVarException):
        blueprint = Blueprint(
            resources=[res.Role(name="role", comment=var.role_comment)],
            vars_spec=[{"name": "role_comment", "type": "string"}],
        )

    blueprint = Blueprint(resources=[res.Role(name="role", comment=var.role_comment)])
    with pytest.raises(MissingVarException):
        blueprint.generate_manifest(session_ctx)


def test_blueprint_vars_in_owner(session_ctx):
    blueprint = Blueprint(
        resources=[res.Schema(name="schema", owner="role_{{ var.role_name }}", database="STATIC_DATABASE")],
        vars={"role_name": "role123"},
    )
    assert blueprint.generate_manifest(session_ctx)


def test_blueprint_sync_resources(session_ctx, remote_state):
    blueprint = Blueprint(
        resources=[res.Role(name="role1")],
        sync_resources=[ResourceType.ROLE],
    )
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    assert len(plan) == 1

    blueprint = Blueprint(sync_resources=["ROLE"])
    assert blueprint._config.sync_resources == [ResourceType.ROLE]
    # Note: sync_resources only affects remote state syncing, not resource validation during add
    # The following validations were expected but are not implemented:
    # - Adding a Database when sync_resources=[ROLE] should raise InvalidResourceException
    # - Creating a Blueprint with Role when sync_resources=[DATABASE] should raise InvalidResourceException


def test_merge_account_scoped_resources():
    resources = [
        res.Database(name="DB1"),
        ResourcePointer(name="DB1", resource_type=ResourceType.DATABASE),
    ]
    merged = _merge_pointers(resources)
    assert len(merged) == 1
    assert isinstance(merged[0], res.Database)
    assert merged[0].name == "DB1"

    resources = [
        res.Database(name="DB1"),
        res.Database(name="DB2"),
    ]
    merged = _merge_pointers(resources)
    assert len(merged) == 2


def test_merge_account_scoped_resources_fail():
    resources = [
        res.Database(name="DB1"),
        res.Database(name="DB1", comment="namespace conflict"),
    ]
    with pytest.raises(DuplicateResourceException):
        _merge_pointers(resources)


def test_blueprint_edition_checks(session_ctx, remote_state):
    session_ctx = deepcopy(session_ctx)
    session_ctx["account_edition"] = AccountEdition.STANDARD

    blueprint = Blueprint(resources=[res.Database(name="DB1"), res.Tag(name="TAG1")])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    with pytest.raises(NonConformingPlanException):
        blueprint._raise_for_nonconforming_plan(session_ctx, plan)

    blueprint = Blueprint(resources=[res.Warehouse(name="WH", min_cluster_count=2)])
    with pytest.raises(WrongEditionException):
        blueprint.generate_manifest(session_ctx)

    blueprint = Blueprint(resources=[res.Warehouse(name="WH", min_cluster_count=1)])
    assert blueprint.generate_manifest(session_ctx)

    blueprint = Blueprint(resources=[res.Warehouse(name="WH")])
    assert blueprint.generate_manifest(session_ctx)


def test_blueprint_warehouse_scaling_policy_doesnt_render_in_standard_edition(session_ctx, remote_state):
    session_ctx = deepcopy(session_ctx)
    session_ctx["account_edition"] = AccountEdition.STANDARD
    wh = res.Warehouse(name="WH", warehouse_size="XSMALL")
    blueprint = Blueprint(resources=[wh])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    assert len(plan) == 1
    assert isinstance(plan[0], CreateResource)
    sql = flatten_sql_commands(compile_plan_to_sql(session_ctx, plan))
    assert len(sql) == 3
    assert sql[0] == "USE SECONDARY ROLES ALL"
    assert sql[1] == "USE ROLE SYSADMIN"
    assert sql[2].startswith("CREATE WAREHOUSE WH")
    assert "scaling_policy" not in sql[2]


def test_blueprint_scope_config():

    bc = BlueprintConfig(
        scope=BlueprintScope.DATABASE,
        database=ResourceName("foo"),
    )
    assert bc

    with pytest.raises(ValueError):
        BlueprintConfig(
            scope=BlueprintScope.DATABASE,
            schema=ResourceName("bar"),
        )

    with pytest.raises(ValueError):
        BlueprintConfig(
            scope=BlueprintScope.ACCOUNT,
            database=ResourceName("foo"),
        )

    with pytest.raises(ValueError):
        BlueprintConfig(
            scope=BlueprintScope.ACCOUNT,
            schema=ResourceName("bar"),
        )

    with pytest.raises(ValueError):
        BlueprintConfig(
            scope=BlueprintScope.ACCOUNT,
            database=ResourceName("foo"),
            schema=ResourceName("bar"),
        )


def test_blueprint_scope(session_ctx, remote_state):

    blueprint = Blueprint(resources=[res.Database(name="DB1")], scope=BlueprintScope.DATABASE)
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    assert len(plan) == 1

    blueprint = Blueprint(resources=[res.Role(name="ROLE1")], scope=BlueprintScope.DATABASE)
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    with pytest.raises(NonConformingPlanException):
        blueprint._raise_for_nonconforming_plan(session_ctx, plan)

    remote_state = {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
        parse_URN("urn::ABCD123:database/DB1"): {"owner": "SYSADMIN"},
        parse_URN("urn::ABCD123:schema/DB1.PUBLIC"): {"owner": "SYSADMIN"},
    }

    blueprint = Blueprint(
        resources=[
            res.Schema(name="SCHEMA1"),
            res.Task(name="TASK1"),
        ],
        scope=BlueprintScope.DATABASE,
        database="DB1",
    )
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    assert len(plan) == 2

    blueprint = Blueprint(resources=[res.Database(name="DB2")], scope=BlueprintScope.SCHEMA)
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    with pytest.raises(NonConformingPlanException):
        blueprint._raise_for_nonconforming_plan(session_ctx, plan)


def test_blueprint_plan_scope_stubbing(session_ctx):
    remote_state = {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
        parse_URN("urn::ABCD123:database/DB1"): {"owner": "SYSADMIN"},
        parse_URN("urn::ABCD123:schema/DB1.PUBLIC"): {"owner": "SYSADMIN"},
    }

    blueprint = Blueprint(
        resources=[res.Task(name="TASK1")],
        scope=BlueprintScope.SCHEMA,
        database="DB1",
        schema="PUBLIC",
    )
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    assert len(plan) == 1

    remote_state = {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
        parse_URN("urn::ABCD123:database/DB1"): {"owner": "SYSADMIN"},
        parse_URN("urn::ABCD123:schema/DB1.PUBLIC"): {"owner": "SYSADMIN"},
        parse_URN("urn::ABCD123:schema/DB1.ANOTHER_SCHEMA"): {"owner": "SYSADMIN"},
    }

    blueprint = Blueprint(
        resources=[res.Task(name="TASK1")],
        scope=BlueprintScope.SCHEMA,
        database="DB1",
        schema="ANOTHER_SCHEMA",
    )
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    assert len(plan) == 1

    remote_state = {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
        parse_URN("urn::ABCD123:database/DB1"): {"owner": "SYSADMIN"},
        parse_URN("urn::ABCD123:schema/DB1.PUBLIC"): {"owner": "SYSADMIN"},
    }

    blueprint = Blueprint(
        resources=[res.Schema(name="A_THIRD_SCHEMA"), res.Task(name="TASK1")],
        scope=BlueprintScope.SCHEMA,
        database="DB1",
        schema="A_THIRD_SCHEMA",
    )
    manifest = blueprint.generate_manifest(session_ctx)
    plan = diff(remote_state, manifest)
    assert len(plan) == 2


def test_resource_type_needs_params(session_ctx):
    """Test that resource_type_needs_params correctly identifies when param fetching is needed.

    Parameter fields (like max_data_extension_time_in_days) now default to None,
    meaning "inherit from parent". The optimization skips SHOW PARAMETERS when
    no resource explicitly sets these fields.
    """
    from snowcap.blueprint import (
        Blueprint,
        resource_type_needs_params,
        schema_urn_needs_params,
        databases_with_param_fields,
    )
    from snowcap.identifiers import FQN, ResourceName

    # Schema type-level check always returns True (delegates to per-URN check)
    blueprint = Blueprint(
        resources=[
            res.Schema(name="MY_SCHEMA", database="MY_DB", owner="SYSADMIN"),
        ]
    )
    manifest = blueprint.generate_manifest(session_ctx)
    assert resource_type_needs_params(ResourceType.SCHEMA, manifest) is True  # Delegates to per-URN

    # Test per-URN schema check: schema without explicit param fields - should NOT need params
    db_with_params = databases_with_param_fields(manifest)
    schema_urn = URN(
        resource_type=ResourceType.SCHEMA,
        fqn=FQN(ResourceName("MY_SCHEMA"), database=ResourceName("MY_DB")),
        account_locator=session_ctx["account_locator"],
    )
    assert schema_urn_needs_params(schema_urn, manifest, db_with_params) is False

    # Schema with explicit default_ddl_collation - per-URN check SHOULD need params
    blueprint = Blueprint(
        resources=[
            res.Schema(
                name="MY_SCHEMA",
                database="MY_DB",
                owner="SYSADMIN",
                default_ddl_collation="en_US",
            ),
        ]
    )
    manifest = blueprint.generate_manifest(session_ctx)
    db_with_params = databases_with_param_fields(manifest)
    assert schema_urn_needs_params(schema_urn, manifest, db_with_params) is True

    # Schema with explicit max_data_extension_time_in_days - per-URN check SHOULD need params
    blueprint = Blueprint(
        resources=[
            res.Schema(
                name="MY_SCHEMA",
                database="MY_DB",
                owner="SYSADMIN",
                max_data_extension_time_in_days=28,
            ),
        ]
    )
    manifest = blueprint.generate_manifest(session_ctx)
    db_with_params = databases_with_param_fields(manifest)
    assert schema_urn_needs_params(schema_urn, manifest, db_with_params) is True

    # Database without explicit parameter fields - should NOT need params
    blueprint = Blueprint(
        resources=[
            res.Database(name="MY_DB", owner="SYSADMIN"),
        ]
    )
    manifest = blueprint.generate_manifest(session_ctx)
    assert resource_type_needs_params(ResourceType.DATABASE, manifest) is False

    # Database with explicit max_data_extension_time_in_days - SHOULD need params
    blueprint = Blueprint(
        resources=[
            res.Database(name="MY_DB", owner="SYSADMIN", max_data_extension_time_in_days=7),
        ]
    )
    manifest = blueprint.generate_manifest(session_ctx)
    assert resource_type_needs_params(ResourceType.DATABASE, manifest) is True

    # Database with params + schema without params - SCHEMA type-level check returns True
    # (delegates to per-URN check via schema_urn_needs_params)
    blueprint = Blueprint(
        resources=[
            res.Database(name="MY_DB", owner="SYSADMIN", max_data_extension_time_in_days=7),
            res.Schema(name="MY_SCHEMA", database="MY_DB", owner="SYSADMIN"),
        ]
    )
    manifest = blueprint.generate_manifest(session_ctx)
    assert resource_type_needs_params(ResourceType.DATABASE, manifest) is True
    assert resource_type_needs_params(ResourceType.SCHEMA, manifest) is True  # Delegates to per-URN

    # Per-URN check: PUBLIC schema needs params when database has params (inheritance)
    db_with_params = databases_with_param_fields(manifest)
    assert "MY_DB" in db_with_params
    public_schema_urn = URN(
        resource_type=ResourceType.SCHEMA,
        fqn=FQN(ResourceName("PUBLIC"), database=ResourceName("MY_DB")),
        account_locator=session_ctx["account_locator"],
    )
    assert schema_urn_needs_params(public_schema_urn, manifest, db_with_params) is True

    # Per-URN check: Non-PUBLIC schema without params does NOT need params
    other_schema_urn = URN(
        resource_type=ResourceType.SCHEMA,
        fqn=FQN(ResourceName("MY_SCHEMA"), database=ResourceName("MY_DB")),
        account_locator=session_ctx["account_locator"],
    )
    assert schema_urn_needs_params(other_schema_urn, manifest, db_with_params) is False

    # Empty manifest - SCHEMA type-level check still returns True (delegates to per-URN)
    blueprint = Blueprint(resources=[])
    manifest = blueprint.generate_manifest(session_ctx)
    assert resource_type_needs_params(ResourceType.SCHEMA, manifest) is True  # Delegates to per-URN

    # Roles have no PARAMETER_FIELDS entry - should always return True (no optimization)
    blueprint = Blueprint(
        resources=[res.Role(name="MY_ROLE")],
    )
    manifest = blueprint.generate_manifest(session_ctx)
    assert resource_type_needs_params(ResourceType.ROLE, manifest) is True
